from policyengine_us.model_api import *


class meets_wic_categorical_eligibility(Variable):
    value_type = bool
    entity = Person
    definition_period = MONTH
    documentation = (
        "Meets the program participation eligibility criteria for WIC"
    )
    label = "Meets WIC categorical (program participation) eligibility"
    reference = "https://www.law.cornell.edu/uscode/text/42/1786#d_2_A"

    def formula(person, period, parameters):
        spm_unit = person.spm_unit
        # https://www.law.cornell.edu/uscode/text/42/1786#d_2_A_ii
        receives_snap_or_tanf = add(spm_unit, period, ["snap", "tanf"]) > 0
        # https://www.law.cornell.edu/uscode/text/42/1786#d_2_A_iii
        receives_medicaid = person("medicaid_enrolled", period)
        # "is a member of a family in which a pregnant woman or an infant receives [Medicaid]."
        pregnant = person("is_pregnant", period)
        wic_category = person("wic_category", period)
        infant = wic_category == wic_category.possible_values.INFANT
        pregnant_or_infant_medicaid_in_spmu = spm_unit.any(
            receives_medicaid & (pregnant | infant)
        )
        return (
            receives_snap_or_tanf
            | receives_medicaid
            | pregnant_or_infant_medicaid_in_spmu
        )



from policyengine_us.model_api import *


class is_chip_fcep_eligible_person(Variable):
    value_type = bool
    entity = Person
    label = "Pregnant person eligible for CHIP through FCEP"
    documentation = "Determines if a pregnant person is eligible for the Children's Health Insurance Program through the Family Coverage Extension Program"
    definition_period = YEAR
    reference = (
        "https://www.kff.org/affordable-care-act/state-indicator/medicaid-and-chip-income-eligibility-limits-for-pregnant-women-as-a-percent-of-the-federal-poverty-level",
    )

    def formula(person, period, parameters):
        # Get state code
        state_code = person.household("state_code", period)

        # Check pregnancy status
        is_pregnant = person("is_pregnant", period)

        # Check if state offers FCEP program for pregnant women
        p = parameters(period).gov.hhs.chip.fcep
        income_limit = p.income_limit[state_code]

        state_has_fcep = income_limit > 0

        # Check immigration status eligibility
        istatus = person("immigration_status", period)
        undocumented = istatus == istatus.possible_values.UNDOCUMENTED
        immigration_eligible = ~undocumented

        # Check income eligibility
        # CHIP is for pregnant women who make too much for Medicaid but below CHIP limits
        # First, check if not eligible for Medicaid
        medicaid_eligible = person("is_medicaid_eligible", period)

        # Check if family income is below CHIP threshold
        # Use medicaid_income_level as the income measure
        income_ratio = person("medicaid_income_level", period)
        income_eligible = income_ratio <= income_limit

        return (
            is_pregnant
            & state_has_fcep
            & immigration_eligible
            & ~medicaid_eligible
            & income_eligible
        )



from policyengine_us.model_api import *


class nm_low_and_middle_income_exemption(Variable):
    value_type = float
    entity = TaxUnit
    label = "New Mexico low- and middle-income exemption"
    defined_for = StateCode.NM
    unit = USD
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.nm.tax.income.exemptions.low_and_middle_income
        filing_status = tax_unit("filing_status", period)
        agi = tax_unit("adjusted_gross_income", period)

        # Calculate the eligibility based on AGI and filing status
        eligible = agi <= p.income_limit[filing_status]

        # Calculate the exemption amount based on AGI and filing status
        reduction_threshold = p.reduction.income_threshold[filing_status]
        excess = max_(agi - reduction_threshold, 0)
        reduction_rate = p.reduction.rate[filing_status]
        reduction_amount = excess * reduction_rate
        exemption_amount = p.max_amount - reduction_amount

        # Multiply the exemption amount by the number of exemptions
        exemptions_count = tax_unit("exemptions_count", period)
        return eligible * exemption_amount * exemptions_count



from policyengine_us.model_api import *


class dc_gac_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Eligible for DC General Assistance for Children (GAC)"
    definition_period = MONTH
    reference = (
        "https://code.dccouncil.gov/us/dc/council/code/sections/4-205.05a#(c)"
    )
    defined_for = "dc_gac_income_eligible"

    def formula(spm_unit, period, parameters):
        has_eligible_child = (
            add(spm_unit, period, ["dc_gac_eligible_child"]) > 0
        )
        resources_eligible = spm_unit("dc_tanf_resources_eligible", period)

        return has_eligible_child & resources_eligible



from policyengine_us.model_api import *


class k12_tuition_and_fees(Variable):
    value_type = float
    entity = TaxUnit
    label = "K-12 Tuition and fees (from Form 8917)"
    unit = USD
    definition_period = YEAR



from policyengine_us.model_api import *


class az_increased_excise_tax_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Arizona Increased Excise Tax Credit"
    unit = USD
    definition_period = YEAR
    defined_for = "az_increased_excise_tax_credit_eligible"
    reference = "https://www.azleg.gov/viewdocument/?docName=https://www.azleg.gov/ars/43/01072-01.htm"

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.az.tax.income.credits.increased_excise
        # The increased excise tax credit is allowed for each person that a
        # personal or dependent exemption can be claimed for
        tax_unit_size = tax_unit("tax_unit_size", period)
        uncapped_credit = tax_unit_size * p.amount
        return min_(uncapped_credit, p.max_amount)



from policyengine_us.model_api import *


class in_eitc(Variable):
    value_type = float
    entity = TaxUnit
    label = "Indiana earned income tax credit"
    unit = USD
    definition_period = YEAR
    reference = "https://iga.in.gov/laws/2021/ic/titles/6#6-3.1-21"
    defined_for = "in_eitc_eligible"

    def formula(tax_unit, period, parameters):
        ip = parameters(period).gov.states["in"].tax.income.credits
        if not ip.earned_income.decoupled:
            federal_eitc = tax_unit("eitc", period)
            return federal_eitc * ip.earned_income.match_rate
        # if Indiana EITC is decoupled from federal EITC
        fp = parameters(period).gov.irs.credits
        # ... cap child count
        kid_cap = ip.earned_income.max_children
        kids = min_(kid_cap, tax_unit("eitc_child_count", period))
        # ... specify decoupled parameter values
        maximum = fp.eitc.max.calc(kids)
        pi_rate = fp.eitc.phase_in_rate.calc(kids)
        po_start = fp.eitc.phase_out.start.calc(kids)  # no JOINT bonus
        po_rate = fp.eitc.phase_out.rate.calc(kids)
        if str(period) == "2021":
            # ... additional decoupling of parameters for childless taxpayers
            maximum0 = ip.earned_income.childless.maximum
            pi_rate0 = ip.earned_income.childless.phase_in_rate
            po_start0 = ip.earned_income.childless.phase_out_start
            po_rate0 = ip.earned_income.childless.phase_out_rate
            # ... integrate in childless parameters
            maximum = where(kids == 0, maximum0, maximum)
            pi_rate = where(kids == 0, pi_rate0, pi_rate)
            po_start = where(kids == 0, po_start0, po_start)
            po_rate = where(kids == 0, po_rate0, po_rate)
        # ... calculate eitc phase-in amount
        earnings = tax_unit("filer_adjusted_earnings", period)
        phase_in_amount = min_(earnings * pi_rate, maximum)
        # ... calculate eitc reduction
        federal_agi = tax_unit("adjusted_gross_income", period)
        higher_income = max_(earnings, federal_agi)
        reduction = po_rate * max_(0, higher_income - po_start)
        # ... calculate decoupled eitc amount
        amount = min_(phase_in_amount, max_(0, maximum - reduction))
        # ... match decoupled eitc amount to get Indiana eitc
        return amount * ip.earned_income.match_rate



from policyengine_us.model_api import *


class dividend_income(Variable):
    value_type = float
    entity = Person
    label = "ordinary dividend income"
    documentation = "Qualified and non-qualified dividends"
    unit = USD
    definition_period = YEAR
    adds = [
        "qualified_dividend_income",
        "non_qualified_dividend_income",
    ]



from policyengine_us.model_api import *


class il_is_exemption_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Whether this tax unit is eligible for any exemptions"
    unit = USD
    definition_period = YEAR

    defined_for = StateCode.IL

    def formula(tax_unit, period, parameters):
        filing_status = tax_unit("filing_status", period)
        cap = parameters(
            period
        ).gov.states.il.tax.income.exemption.income_limit[filing_status]
        return tax_unit("adjusted_gross_income", period) <= cap



from policyengine_us.model_api import *


class SSICategory(Enum):
    AGED = "Aged"
    BLIND = "Blind"
    DISABLED = "Disabled"
    NONE = "None"


class ssi_category(Variable):
    value_type = Enum
    entity = Person
    label = "SSI category"
    definition_period = YEAR
    possible_values = SSICategory
    default_value = SSICategory.NONE

    def formula(person, period, parameters):
        is_blind = person("is_blind", period)
        is_aged = person("is_ssi_aged", period)
        is_disabled = person("is_ssi_disabled", period)

        # If multiple are true, pick whichever is your policy (often Blind overrides Aged, or we do a priority).
        # Here's a simple approach: blind > aged > disabled, or pick the first that's true:
        return select(
            [is_blind, is_aged, is_disabled],
            [SSICategory.BLIND, SSICategory.AGED, SSICategory.DISABLED],
            default=SSICategory.NONE,
        )



from policyengine_us.model_api import *


class mi_homestead_property_tax_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Michigan homestead property tax credit"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/MI-1040CR.pdf#page=2",
        "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf#page=34",
    )
    defined_for = "mi_homestead_property_tax_credit_eligible"

    def formula(tax_unit, period, parameters):
        allowable_credit_amount = tax_unit(
            "mi_homestead_property_tax_credit_pre_alternate_senior_amount",
            period,
        )
        senior_credit_amount = tax_unit(
            "mi_homestead_property_tax_credit_alternate_senior_amount", period
        )
        return max_(allowable_credit_amount, senior_credit_amount)



from policyengine_us.model_api import *


class mo_business_income_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Missouri business income deduction"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://dor.mo.gov/forms/MO-1040%20Instructions_2023.pdf#page=16",
        "https://revisor.mo.gov/main/OneSection.aspx?section=143.022",
    )
    defined_for = StateCode.MO

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.mo.tax.income.deductions.business_income
        qualified_business_income = add(
            tax_unit, period, ["qualified_business_income"]
        )
        total_qualified_business_income = tax_unit.sum(
            qualified_business_income
        )
        return p.rate * qualified_business_income



from policyengine_us.model_api import *


class recovery_rebate_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Recovery Rebate Credit"
    unit = USD
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/6428"

    adds = ["rrc_cares", "rrc_caa", "rrc_arpa"]



from policyengine_us.model_api import *


class az_property_tax_credit_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Eligible for the Arizona Property Tax Credit"
    definition_period = YEAR
    reference = "https://www.azleg.gov/viewdocument/?docName=https://www.azleg.gov/ars/43/01072.htm"
    defined_for = StateCode.AZ

    def formula(tax_unit, period, parameters):
        person = tax_unit.members
        p = parameters(period).gov.states.az.tax.income.credits.property_tax

        head_or_spouse = person("is_tax_unit_head_or_spouse", period)

        age = person("age", period)
        age_eligible_head_or_spouse = (age >= p.age_threshold) & head_or_spouse

        receives_ssi = person("ssi", period) > 0
        head_or_spouse_receives_ssi = (receives_ssi > 0) & head_or_spouse

        age_or_ssi_eligible = tax_unit.any(
            age_eligible_head_or_spouse | head_or_spouse_receives_ssi
        )

        paid_rent_or_property_tax = (
            add(tax_unit, period, ["rent", "real_estate_taxes"]) > 0
        )

        return age_or_ssi_eligible & paid_rent_or_property_tax



from policyengine_us.model_api import *


class ma_eaedc_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Eligible for the Massachusetts EAEDC"
    definition_period = MONTH
    defined_for = StateCode.MA
    reference = (
        "https://www.law.cornell.edu/regulations/massachusetts/106-CMR-703-010"
    )

    def formula(spm_unit, period, parameters):
        financial_eligible = spm_unit("ma_eaedc_financial_eligible", period)
        non_financial_eligible = spm_unit(
            "ma_eaedc_non_financial_eligible", period
        )
        return financial_eligible & non_financial_eligible



from policyengine_us.model_api import *


class nm_salt_add_back(Variable):
    value_type = float
    entity = TaxUnit
    label = "New Mexico salt addback"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf#page=28",
        "https://law.justia.com/codes/new-mexico/chapter-7/article-2/section-7-2-2/",
        "https://klvg4oyd4j.execute-api.us-west-2.amazonaws.com/prod/PublicFiles/34821a9573ca43e7b06dfad20f5183fd/1afc56af-ea90-4d48-82e5-1f9aeb43255a/PITbook2022.pdf#page=28",
    )
    defined_for = StateCode.NM

    def formula(tax_unit, period, parameters):
        itemizes = tax_unit("tax_unit_itemizes", period)
        filing_status = tax_unit("filing_status", period)
        p = parameters(period).gov.irs.deductions

        # tax/income, federal Schedule A, line 5a. 1
        salt_sales_or_income = tax_unit(
            "state_and_local_sales_or_income_tax", period
        )
        # state_and_local_tax, line 5d. 2
        total_salt = (
            add(tax_unit, period, ["real_estate_taxes"]) + salt_sales_or_income
        )
        # ratio = round(salt_sales_or_income[0] / total_salt[0], 4)
        # ratio. 3
        salt_ratio = np.zeros_like(total_salt)
        mask = total_salt != 0
        salt_ratio[mask] = salt_sales_or_income[mask] / total_salt[mask]

        # line 5e. 4
        salt_cap = p.itemized.salt_and_real_estate.cap[filing_status]
        salt_claimed = min_(salt_cap, total_salt)

        # 5 = salt_claimed * salt_ratio
        # 6 = min(5, 4)
        salt = min_(salt_claimed * salt_ratio, salt_claimed)

        standard_deduction = tax_unit("standard_deduction", period)
        itm_deds_less_salt = tax_unit("itemized_deductions_less_salt", period)
        us_itemized_deductions = itm_deds_less_salt + salt_claimed
        item_deds = max_(us_itemized_deductions - standard_deduction, 0)

        nm_item = min_(salt, item_deds)
        return where(itemizes, nm_item, 0)



from policyengine_us.model_api import *


class md_unmarried_childless_non_refundable_eitc(Variable):
    value_type = float
    entity = TaxUnit
    label = "Maryland unmarried childless non-refundable EITC"
    unit = USD
    definition_period = YEAR
    reference = "https://casetext.com/statute/code-of-maryland/article-tax-general/title-10-income-tax/subtitle-7-income-tax-credits/section-10-704-effective-until-6302023-for-earned-income"  # (c)(3)
    defined_for = "md_qualifies_for_unmarried_childless_eitc"

    def formula(tax_unit, period, parameters):
        # individuals can claim the state eitc even they do not meet the minimum age requirement under the federal credit
        federal_eitc_without_age_minimum = tax_unit(
            "federal_eitc_without_age_minimum", period
        )
        p = parameters(
            period
        ).gov.states.md.tax.income.credits.eitc.non_refundable.unmarried_childless
        match = p.match * federal_eitc_without_age_minimum
        cap = p.max_amount
        return min_(match, cap)



from policyengine_us.model_api import *


class in_section_179_expense_add_back(Variable):
    value_type = float
    entity = TaxUnit
    label = "Indiana Section 179 Expense Add Back"
    definition_period = YEAR
    documentation = "Federal IRC Section 179 expenses less IRC Section 179 expenses if they had been calculated with a $25,000 ceiling."
    reference = "http://iga.in.gov/legislative/laws/2021/ic/titles/006#6-3-1-3.5"  # (a)(17)(A)
    # use federal variables if they are added later
    defined_for = StateCode.IN



from policyengine_us.model_api import *


class self_employment_medicare_tax(Variable):
    value_type = float
    entity = Person
    label = "Self-employment Medicare tax"
    definition_period = YEAR
    unit = USD
    reference = "https://www.law.cornell.edu/uscode/text/26/1401#b_1"

    def formula(person, period, parameters):
        p = parameters(period).gov.irs.self_employment.rate
        return p.medicare * person("taxable_self_employment_income", period)



from policyengine_us.model_api import *


class nj_total_income(Variable):
    value_type = float
    entity = Person
    label = "New Jersey total income by person"
    unit = USD
    definition_period = YEAR
    reference = "https://law.justia.com/codes/new-jersey/2022/title-54/section-54-8a-36/"
    defined_for = StateCode.NJ

    def formula(person, period, parameters):
        agi = person("adjusted_gross_income_person", period)
        nj_additions = person("nj_additions", period)
        nj_subtractions = person("nj_agi_subtractions", period)
        return max_(0, agi + nj_additions - nj_subtractions)



from policyengine_us.model_api import *


class ma_eaedc_financial_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Financial eligible for Massachusetts EAEDC"
    definition_period = MONTH
    defined_for = StateCode.MA
    reference = "https://www.law.cornell.edu/regulations/massachusetts/department-106-CMR/title-106-CMR-704.000"

    def formula(spm_unit, period, parameters):
        assets_eligible = spm_unit("ma_eaedc_assets_limit_eligible", period)
        income_eligible = spm_unit("ma_eaedc_income_eligible", period)

        return assets_eligible & income_eligible



from policyengine_us.model_api import *


class household_refundable_state_tax_credits(Variable):
    value_type = float
    entity = Household
    label = "refundable State income tax credits"
    unit = USD
    definition_period = YEAR
    adds = "gov.household.household_refundable_state_credits"



from policyengine_us.model_api import *


class ma_tafdc_countable_unearned_income(Variable):
    value_type = float
    unit = USD
    entity = SPMUnit
    label = "Massachusetts Temporary Assistance for Families with Dependent Children (TAFDC) countable unearned income"
    definition_period = MONTH
    reference = (
        "https://www.masslegalservices.org/content/62-what-income-counted"
    )
    defined_for = StateCode.MA

    def formula(spm_unit, period, parameters):
        total_unearned_income = add(
            spm_unit, period, ["ma_tcap_gross_unearned_income"]
        )
        child_support_deduction = add(
            spm_unit, period, ["ma_tafdc_child_support_deduction"]
        )
        return max_(0, total_unearned_income - child_support_deduction)



from policyengine_us.model_api import *


class is_ssi_eligible(Variable):
    value_type = bool
    entity = Person
    label = "Is SSI eligible person"
    definition_period = YEAR

    def formula(person, period, parameters):
        abd_person = person("is_ssi_aged_blind_disabled", period)
        meets_resource_test = person("meets_ssi_resource_test", period)
        is_qualified_noncitizen = person("is_ssi_qualified_noncitizen", period)
        immigration_status = person("immigration_status", period)
        is_citizen = (
            immigration_status == immigration_status.possible_values.CITIZEN
        )
        meets_immigration_status = is_qualified_noncitizen | is_citizen

        return abd_person & meets_resource_test & meets_immigration_status



from policyengine_us.model_api import *


class taxsim_age1(Variable):
    value_type = int
    entity = TaxUnit
    label = "Age of first dependent"
    unit = "year"
    documentation = """Age of first dependent. Used for EITC, CTC and CCC. For 1991+ code students between 20 and 23 as 19 to get the EITC calculation correct. Code infants as "1". [For compatibiity with taxsim32, dep13-dep18 are accepted and have priority over age1-age3]. If niether dep19 or age1 are present in an uploaded file than depx is used for the number of child eligible for the EIC, CTC and CDCC."""
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        person = tax_unit.members
        is_dependent = person("is_tax_unit_dependent", period)
        age = person("age", period)
        dependent_rank = person.get_rank(tax_unit, age, is_dependent)
        is_first_dependent = dependent_rank == 0
        return tax_unit.sum(age * is_first_dependent)



from policyengine_us.model_api import *


class amt_excluded_deductions(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "AMT taxable income excluded deductions"
    unit = USD
    reference = "https://www.law.cornell.edu/uscode/text/26/55#b_2"

    def formula(tax_unit, period, parameters):
        itemizing = tax_unit("tax_unit_itemizes", period)
        standard_deduction = tax_unit("standard_deduction", period)
        p = parameters(period).gov.irs.income.amt
        itemized_deductions_add_back = add(
            tax_unit, period, p.itemized_deductions_add_back
        )
        return where(
            itemizing, itemized_deductions_add_back, standard_deduction
        )



from policyengine_us.model_api import *


class taxable_private_pension_income(Variable):
    value_type = float
    entity = Person
    label = "taxable private pension income"
    unit = USD
    documentation = "Taxable income from non-government employee pensions."
    definition_period = YEAR



from policyengine_us.model_api import *


class mt_standard_deduction_indiv(Variable):
    value_type = float
    entity = Person
    label = (
        "Montana standard deduction when married couples are filing separately"
    )
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MT

    def formula(person, period, parameters):
        # ▸ MT legacy parameter table (only relevant pre-2024)
        p = parameters(period).gov.states.mt.tax.income.deductions.standard

        filing_status = person.tax_unit(
            "state_filing_status_if_married_filing_separately_on_same_return",
            period,
        )

        if p.state_specific_deduction_applies:
            # ── Pre-2024 MT-specific calculation ────────────────────────────────
            agi = person("mt_agi", period)
            floor = p.floor[filing_status]
            cap = p.cap[filing_status]
            uncapped = p.rate * agi
            deduction_amount = max_(min_(uncapped, cap), floor)

        else:
            # ── 2024-onward: apply the *federal* standard-deduction variable ──

            deduction_amount = person.tax_unit("standard_deduction", period)

        is_head_or_spouse = person("is_tax_unit_head_or_spouse", period)
        return is_head_or_spouse * deduction_amount



from policyengine_us.model_api import *


class ok_federal_eitc(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Federal earned income credit for the Oklahoma EITC computation"
    reference = (
        "https://law.justia.com/codes/oklahoma/title-68/section-68-2357-43/"
    )
    unit = USD
    defined_for = "ok_federal_eitc_eligible"

    def formula(tax_unit, period, parameters):
        takes_up_eitc = tax_unit("takes_up_eitc", period)
        maximum = tax_unit("ok_federal_eitc_maximum", period)
        phased_in = tax_unit("ok_federal_eitc_phased_in", period)
        reduction = tax_unit("ok_federal_eitc_reduction", period)
        limitation = max_(0, maximum - reduction)
        return min_(phased_in, limitation) * takes_up_eitc



from policyengine_us.model_api import *


class snap_max_allotment(Variable):
    value_type = float
    entity = SPMUnit
    definition_period = MONTH
    documentation = "Maximum SNAP allotment for SPM unit, based on the state group and household size."
    label = "SNAP maximum allotment"
    unit = USD

    def formula(spm_unit, period, parameters):
        max_allotments = parameters(period).gov.usda.snap.max_allotment
        MAX_HOUSEHOLD_SIZE_UNDER_MAIN = 8
        snap_region = spm_unit.household("snap_region_str", period)
        household_size = min_(
            spm_unit("snap_unit_size", period), MAX_HOUSEHOLD_SIZE_UNDER_MAIN
        )
        additional_members = max_(
            0,
            spm_unit("snap_unit_size", period) - MAX_HOUSEHOLD_SIZE_UNDER_MAIN,
        )
        main_allotment = max_allotments.main[snap_region][household_size]
        additional_allotment = (
            additional_members * max_allotments.additional[snap_region]
        )
        return main_allotment + additional_allotment



from policyengine_us.model_api import *


class nc_military_retirement_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "North Carolina military retirement deduction"
    unit = USD
    definition_period = YEAR
    defined_for = "nc_military_retirement_deduction_eligible"
    reference = (
        "https://www.ncdor.gov/2022-d-401-individual-income-tax-instructions/open#page=18",
        "https://law.justia.com/codes/north-carolina/chapter-105/article-4/section-105-153-5/",
    )

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.nc.tax.income.deductions.military_retirement
        military_retirement_benefits = add(
            tax_unit, period, ["military_retirement_pay"]
        )
        return military_retirement_benefits * p.fraction



from policyengine_us.model_api import *


class co_ccap_fpg_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Meets Colorado Child Care Assistance Program poverty-based income eligibility test"
    reference = "https://www.sos.state.co.us/CCR/GenerateRulePdf.do?ruleVersionId=11042&fileName=8%20CCR%201403-1#page=19"
    definition_period = MONTH
    # defined_for = StateCode.CO

    def formula(spm_unit, period, parameters):
        household = spm_unit.household
        state_eligible = household("state_code_str", period.this_year) == "CO"
        monthly_gross_income = np.round(
            spm_unit("co_ccap_countable_income", period),
            2,
        )
        year = period.start.year
        if period.start.month >= 10:
            instant_str = f"{year}-10-01"
        else:
            instant_str = f"{year - 1}-10-01"
        p = parameters(instant_str).gov.states.co.ccap
        # Calculate monthly fpg limit, only for counties in Colorado.
        county = household("county_str", period.this_year)
        fpg_rate = np.zeros_like(county, dtype=float)
        mask = state_eligible

        # Current fix for counties not in the dataset.
        try:
            p.entry.fpg_rate[county[mask]]
        except:
            county = np.array(
                ["DENVER_COUNTY_CO"] * len(county),
            )
        if mask.any():
            fpg_rate[mask] = p.entry.fpg_rate[county[mask]]
        # SNAP FPG is monthly.
        fpg = spm_unit("snap_fpg", period)
        fpg_limit = np.round(fpg * fpg_rate, 2)
        meets_income_limit = monthly_gross_income < fpg_limit
        return state_eligible & meets_income_limit



from policyengine_us.model_api import *


class nc_standard_or_itemized_deductions(Variable):
    value_type = float
    entity = TaxUnit
    label = "North Carolina standard or itemized deductions amount"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.ncdor.gov/2021-d-401-individual-income-tax-instructions/open#page=14"
        "https://www.ncdor.gov/2022-d-401-individual-income-tax-instructions/open#page=14"
    )
    defined_for = StateCode.NC

    def formula(tax_unit, period, parameters):
        # From above instructions for Line 11:
        #   You may deduct from federal adjusted gross income either
        #   the N.C. standard deduction or N.C. itemized deductions.
        #   In most cases, your state income tax will be less if you
        #   take the larger of your N.C. itemized deductions or your
        #   N.C. standard deduction.
        stded = tax_unit("nc_standard_deduction", period)
        itded = tax_unit("nc_itemized_deductions", period)
        return max_(stded, itded)



"""
Vermont Medicaid Spending Breakdown by Enrollment Category

This script calculates Vermont's Medicaid spending breakdown by category using
national per capita spending averages and Vermont's enrollment data.

Data Sources:
- EXHIBIT 21. Medicaid Spending by State, Eligibility Group, and Dually Eligible Status | MACPAC
  https://www.macpac.gov/publication/medicaid-spending-by-state-category-and-source-of-funds/

- Spending by enrollment category was created by multiplying total Medicaid spending
  by state by the percent of spending on aged enrollees.
"""

# Vermont's total Medicaid spending (FY 2022)
TOTAL_SPENDING_BILLIONS = 1.7221

# Vermont enrollment by category (in thousands)
VERMONT_ENROLLMENT = {
    "children": 71,
    "new_adult_group": 79,
    "other_adults": 15,
    "disabled": 20,
    "aged": 22,
}

# National per capita spending averages by category (in thousands of dollars)
# Source: MACPAC analysis of Medicaid spending patterns
NATIONAL_PER_CAPITA_THOUSANDS = {
    "children": 3.53851,
    "new_adult_group": 7.05447,
    "other_adults": 5.00694,
    "disabled": 24.73929,
    "aged": 17.69886,
}


def calculate_vermont_spending_breakdown():
    """
    Calculate Vermont's Medicaid spending by category using national per capita rates.

    Methodology:
    1. Calculate hypothetical spending using national per capita rates
    2. Determine adjustment factor to match Vermont's actual total
    3. Apply adjustment factor proportionally to each category

    Returns:
        dict: Spending by category in billions of dollars
    """
    # Step 1: Calculate hypothetical spending using national rates
    hypothetical_spending = {}
    total_hypothetical = 0

    for category, enrollment_thousands in VERMONT_ENROLLMENT.items():
        # Convert enrollment to actual numbers and calculate spending
        enrollment = enrollment_thousands * 1000
        per_capita = (
            NATIONAL_PER_CAPITA_THOUSANDS[category] * 1000
        )  # Convert to dollars

        # Calculate spending in billions
        spending_billions = (enrollment * per_capita) / 1_000_000_000
        hypothetical_spending[category] = spending_billions
        total_hypothetical += spending_billions

    # Step 2: Calculate adjustment factor
    # Vermont's actual spending differs from what national rates would predict
    adjustment_factor = TOTAL_SPENDING_BILLIONS / total_hypothetical

    # Step 3: Apply adjustment factor to get Vermont's actual spending
    vermont_spending = {}

    for category, hypothetical in hypothetical_spending.items():
        # Scale the hypothetical spending to match Vermont's actual total
        vermont_spending[category] = hypothetical * adjustment_factor

    return vermont_spending, adjustment_factor


def get_spending_breakdown_with_metadata():
    """
    Get Vermont's spending breakdown with additional metadata.

    Returns:
        dict: Complete breakdown with amounts, percentages, and methodology notes
    """
    spending_by_category, adjustment_factor = (
        calculate_vermont_spending_breakdown()
    )

    breakdown = {
        "total_billions": TOTAL_SPENDING_BILLIONS,
        "adjustment_factor": adjustment_factor,
        "methodology": (
            f"Spending breakdown reverse-engineered from national per capita averages. "
            f"Applied {adjustment_factor:.4f} adjustment factor to match actual total of ${TOTAL_SPENDING_BILLIONS}B"
        ),
        "categories": {},
    }

    for category, spending_billions in spending_by_category.items():
        breakdown["categories"][category] = {
            "spending_billions": round(spending_billions, 5),
            "spending_millions": round(spending_billions * 1000, 2),
            "percentage": round(
                (spending_billions / TOTAL_SPENDING_BILLIONS) * 100, 1
            ),
        }

    return breakdown


# Calculate Vermont's spending breakdown
if __name__ == "__main__":
    vermont_breakdown = get_spending_breakdown_with_metadata()

    # The calculated values for Vermont's Medicaid spending by category
    VERMONT_SPENDING_BY_CATEGORY = {
        "children": 0.24474,  # $244.74 million (14.2%)
        "new_adult_group": 0.54290,  # $542.90 million (31.5%)
        "other_adults": 0.07316,  # $73.16 million (4.2%)
        "disabled": 0.48199,  # $481.99 million (28.0%)
        "aged": 0.37931,  # $379.31 million (22.0%)
    }



from policyengine_us.model_api import *


class mt_income_tax_before_refundable_credits_unit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Montana income tax before refundable credits"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MT

    def formula(tax_unit, period, parameters):
        income_tax_before_credits_indiv = add(
            tax_unit, period, ["mt_income_tax_before_refundable_credits_indiv"]
        )
        income_tax_before_credits_joint = tax_unit(
            "mt_income_tax_before_refundable_credits_joint", period
        )
        return min_(
            income_tax_before_credits_indiv, income_tax_before_credits_joint
        )



from policyengine_us.model_api import *


class co_military_retirement_subtraction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Colorado military retirement subtraction"
    defined_for = StateCode.CO
    unit = USD
    reference = (
        "https://tax.colorado.gov/sites/tax/files/documents/DR0104AD_2022.pdf#page=1",
        "https://tax.colorado.gov/sites/tax/files/documents/DR_104_Book_2022.pdf#page=12",
        "https://law.justia.com/codes/colorado/2022/title-39/article-22/part-1/section-39-22-104/",
        # C.R.S. 39-22-104(4)(y)(I)
    )
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.co.tax.income.subtractions.military_retirement
        person = tax_unit.members
        age = person("age", period)
        head = person("is_tax_unit_head", period)
        spouse = person("is_tax_unit_spouse", period)
        head_or_spouse = head | spouse
        age_eligible = age < p.age_threshold
        eligible = head_or_spouse * age_eligible
        military_retirement_pay = person("military_retirement_pay", period)
        capped_military_retirement_pay = min_(
            military_retirement_pay * eligible, p.max_amount
        )
        return tax_unit.sum(capped_military_retirement_pay)



from policyengine_us.model_api import *


class mt_agi(Variable):
    value_type = float
    entity = Person
    label = "Montana Adjusted Gross Income for each individual"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MT

    def formula(person, period, parameters):
        agi = person("adjusted_gross_income_person", period)
        additions = person("mt_additions", period)
        subtractions = person("mt_subtractions", period)
        reduced_agi = max_(agi + additions - subtractions, 0)
        # Montana taxable social security benefits can be either addition or subtraction
        # if mt_taxable_social security - taxable_social_security > 0, then addition, else subtraction

        p = parameters(period).gov.states.mt.tax.income.social_security
        if p.applies:
            # 2023 and before: apply lines 21-24 adjustment for social security
            taxable_ss = person("taxable_social_security", period)
            mt_taxable_ss = person("mt_taxable_social_security", period)
            adjusted_mt_ss_difference = mt_taxable_ss - taxable_ss
            tax_unit_mt_agi = max_(reduced_agi + adjusted_mt_ss_difference, 0)
        else:
            # 2024 and after: no longer apply the social security adjustment
            tax_unit_mt_agi = reduced_agi

        # in all years
        # allocate any dependent net_income to tax unit head
        is_dependent = person("is_tax_unit_dependent", period)
        sum_dep_net_income = person.tax_unit.sum(
            is_dependent * tax_unit_mt_agi
        )
        is_head = person("is_tax_unit_head", period)
        return ~is_dependent * tax_unit_mt_agi + is_head * sum_dep_net_income



from policyengine_us.model_api import *


class hi_food_excise_credit_minor_child_amount(Variable):
    value_type = float
    entity = TaxUnit
    label = "Minor child amount for the Hawaii Food/Excise Tax Credit"
    definition_period = YEAR
    defined_for = StateCode.HI

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.hi.tax.income.credits.food_excise_tax

        return p.minor_child.amount * tax_unit(
            "hi_food_excise_credit_minor_child_count", period
        )



from policyengine_us.model_api import *


class is_adult_for_medicaid_nfc(Variable):
    value_type = bool
    entity = Person
    label = "Medicaid adult non-financial criteria"
    definition_period = YEAR

    def formula(person, period, parameters):
        ma = parameters(period).gov.hhs.medicaid.eligibility.categories.adult
        age = person("age", period)
        return ma.age_range.calc(age)



from policyengine_us.model_api import *


class md_standard_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "MD standard deduction"
    unit = USD
    definition_period = YEAR
    reference = [
        "https://govt.westlaw.com/mdc/Document/NC8EB19606F6911E8A99BCF2C90B83D38?viewType=FullText&originationContext=documenttoc&transitionType=CategoryPageItem&contextData=(sc.Default)#co_anchor_I552E3B107DF711ECA8F2FF3A9E62BB69",
        "https://mgaleg.maryland.gov/2025RS/Chapters_noln/CH_604_hb0352e.pdf#page=165",  # Maryland House Bill 352 - Budget Reconciliation and Financing Act of 2025
    ]
    defined_for = StateCode.MD

    def formula(tax_unit, period, parameters):
        filing_status = tax_unit("filing_status", period)
        p = parameters(period).gov.states.md.tax.income.deductions.standard

        # Use flat amounts when applicable
        if p.flat_deduction.applies:
            return p.flat_deduction.amount[filing_status]

        # For years when flat deduction doesn't apply, use the old formula:
        # standard deduction is a percentage of AGI that
        # is bounded by a min/max by filing status.
        md_agi = tax_unit("md_agi", period)
        return np.clip(
            p.rate * md_agi, p.min[filing_status], p.max[filing_status]
        )



from policyengine_us.model_api import *


class commodity_supplemental_food_program_eligible(Variable):
    value_type = bool
    entity = Person
    definition_period = YEAR
    label = "Commodity Supplemental Food Program eligible"

    def formula(person, period, parameters):
        p = parameters(period).gov.usda.csfp
        # CFR defines CSFP income similarly to WIC income.
        # Assume resources are counted at the SPM unit level.
        fpg = person.spm_unit("school_meal_fpg_ratio", period)
        age = person("age", period)

        age_eligible = age >= p.min_age
        income_eligible = fpg <= p.fpg_limit

        return age_eligible & income_eligible



from policyengine_us.model_api import *


class ga_investment_in_529_plan_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Georgia investment in 529 plan deduction"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://dor.georgia.gov/document/document/2022-it-511-individual-income-tax-booklet"
        "https://casetext.com/regulation/georgia-administrative-code/department-560-rules-of-department-of-revenue/chapter-560-7-income-tax-division/subject-560-7-4-net-taxable-income-individual/rule-560-7-4-04-procedures-governing-the-georgia-higher-education-savings-plan"
    )
    defined_for = StateCode.GA



from policyengine_us.model_api import *


class il_tanf_non_financial_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Eligible for Illinois Temporary Assistance for Needy Families (TANF) due to non financial requirements"
    definition_period = MONTH
    reference = "https://www.dhs.state.il.us/page.aspx?item=30358"
    defined_for = StateCode.IL

    def formula(spm_unit, period, parameters):
        demographic_eligible = (
            add(spm_unit, period, ["il_tanf_demographic_eligible_person"]) > 0
        )
        immigration_status_eligible = (
            add(
                spm_unit,
                period,
                ["il_tanf_immigration_status_eligible_person"],
            )
            > 0
        )
        return demographic_eligible & immigration_status_eligible



from policyengine_us.model_api import *


class or_non_refundable_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "OR uncapped non-refundable tax credits"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.OR

    adds = "gov.states.or.tax.income.credits.nonrefundable"



from policyengine_us.model_api import *


class cdcc_relevant_expenses(Variable):
    value_type = float
    entity = TaxUnit
    label = "CDCC-relevant care expenses"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.law.cornell.edu/uscode/text/26/21#c",
        "https://www.law.cornell.edu/uscode/text/26/21#d_1",
    )

    def formula(tax_unit, period, parameters):
        expenses = tax_unit("tax_unit_childcare_expenses", period)
        cdcc_limit = tax_unit("cdcc_limit", period)
        eligible_capped_expenses = min_(expenses, cdcc_limit)
        # cap further to the lowest earnings between the taxpayer and spouse
        return min_(
            eligible_capped_expenses,
            tax_unit("min_head_spouse_earned", period),
        )



from policyengine_us.model_api import *


class ne_cdcc_refundable(Variable):
    value_type = float
    entity = TaxUnit
    label = "Nebraska refundable cdcc"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://revenue.nebraska.gov/files/doc/tax-forms/2021/f_2441n.pdf"
        "https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/Form_2441N_Ne_Child_and_Dependent_Care_Expenses_8-618-2022_final_2.pdf"
    )
    defined_for = "ne_cdcc_refundable_eligible"

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.ne.tax.income.credits.cdcc.refundable
        us_agi = tax_unit("adjusted_gross_income", period)
        max_match_percentage = p.match
        reduction_start = p.reduction.start
        increment = p.reduction.increment
        reduction_per_increment = p.reduction.amount
        excess = max_(us_agi - reduction_start, 0)
        increments = np.ceil(excess / increment)
        total_match_reduction_amount = increments * reduction_per_increment
        match = max_(max_match_percentage - total_match_reduction_amount, 0)
        # Nebraska matches the potential federal credit
        us_cdcc = tax_unit("cdcc_potential", period)
        return us_cdcc * match



from policyengine_us.model_api import *


def create_capital_gains_tax_increase() -> Reform:
    class capital_gains_tax(Variable):
        value_type = float
        entity = TaxUnit
        label = "Maximum income tax after capital gains tax"
        unit = USD
        definition_period = YEAR

        def formula(tax_unit, period, parameters):
            net_cg = tax_unit("net_capital_gain", period)
            taxable_income = tax_unit("taxable_income", period)
            adjusted_net_cg = min_(
                tax_unit("adjusted_net_capital_gain", period),
                taxable_income,
            )  # ANCG is referred to in all cases as ANCG or taxable income if less.

            cg = parameters(period).gov.irs.capital_gains

            excluded_cg = tax_unit(
                "capital_gains_excluded_from_taxable_income", period
            )
            non_cg_taxable_income = max_(0, taxable_income - excluded_cg)
            income_less_ancg = max_(0, taxable_income - adjusted_net_cg)

            filing_status = tax_unit("filing_status", period)

            first_threshold = cg.brackets.thresholds["1"][filing_status]
            second_threshold = cg.brackets.thresholds["2"][filing_status]

            income_ordinarily_under_second_rate = clip(
                taxable_income, 0, first_threshold
            )

            income_ordinarily_under_third_rate = clip(
                taxable_income, 0, second_threshold
            )

            # The 2025 Biden Budget taxes the excess of long-term capital gains and qualified dividends
            # over $1 million as ordinary income.
            # We apply this only to ANCG, not Unrecaptured Section 1250 Gain or the 28% rate CG.

            p_reform = parameters(
                period
            ).gov.contrib.biden.budget_2025.capital_gains
            income_threshold = p_reform.income_threshold[filing_status]
            excess_income = max_(0, taxable_income - income_threshold)

            cg_in_top_bracket = min_(adjusted_net_cg, excess_income)

            income_tax = parameters(period).gov.irs.income.bracket.rates["7"]

            new_cg_tax = cg_in_top_bracket * income_tax

            cg_in_first_bracket = max_(
                0, income_ordinarily_under_second_rate - income_less_ancg
            )

            cg_in_first_bracket_below_top_bracket = max_(
                cg_in_first_bracket - cg_in_top_bracket, 0
            )

            cg_in_second_bracket = min_(
                max_(
                    0, adjusted_net_cg - cg_in_first_bracket_below_top_bracket
                ),
                max_(
                    0,
                    income_ordinarily_under_third_rate
                    - (
                        non_cg_taxable_income
                        + cg_in_first_bracket_below_top_bracket
                    ),
                ),
            )

            cg_in_second_bracket_below_top_bracket = max_(
                cg_in_second_bracket - cg_in_top_bracket, 0
            )

            cg_in_third_bracket = max_(
                adjusted_net_cg
                - cg_in_first_bracket_below_top_bracket
                - cg_in_second_bracket_below_top_bracket,
                0,
            )

            cg_in_third_bracket_below_top_bracket = max_(
                cg_in_third_bracket - cg_in_top_bracket, 0
            )

            main_cg_tax = (
                cg_in_first_bracket_below_top_bracket * cg.brackets.rates["1"]
                + cg_in_second_bracket_below_top_bracket
                * cg.brackets.rates["2"]
                + cg_in_third_bracket_below_top_bracket
                * cg.brackets.rates["3"]
                + new_cg_tax
            )

            unrecaptured_s_1250_gain = tax_unit(
                "unrecaptured_section_1250_gain", period
            )

            qualified_dividends = add(
                tax_unit, period, ["qualified_dividend_income"]
            )

            max_taxable_unrecaptured_gain = min_(
                unrecaptured_s_1250_gain,
                max_(0, net_cg - qualified_dividends),
            )
            unrecaptured_gain_deduction = max_(
                non_cg_taxable_income + net_cg - taxable_income,
                0,
            )
            taxable_unrecaptured_gain = max_(
                max_taxable_unrecaptured_gain - unrecaptured_gain_deduction,
                0,
            )

            unrecaptured_gain_tax = (
                cg.unrecaptured_s_1250_rate * taxable_unrecaptured_gain
            )

            remaining_cg_tax = (
                tax_unit("capital_gains_28_percent_rate_gain", period)
                * cg.other_cg_rate
            )
            return main_cg_tax + unrecaptured_gain_tax + remaining_cg_tax

    class reform(Reform):
        def apply(self):
            self.update_variable(capital_gains_tax)

    return reform


def create_capital_gains_tax_increase_reform(
    parameters, period, bypass: bool = False
):
    if bypass:
        return create_capital_gains_tax_increase()

    p = parameters(period).gov.contrib.biden.budget_2025.capital_gains

    if p.active:
        return create_capital_gains_tax_increase()
    else:
        return None


capital_gains_tax_increase = create_capital_gains_tax_increase_reform(
    None, None, bypass=True
)



from policyengine_us.model_api import *


class mi_standard_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Michigan standard deduction"
    unit = USD
    definition_period = YEAR
    documentation = "Michigan standard deduction of qualifying age."
    reference = (
        "http://legislature.mi.gov/doc.aspx?mcl-206-30",
        "https://www.michigan.gov/taxes/iit/retirement-and-pension-benefits/michigan-standard-deduction",
        "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/BOOK_MI-1040.pdf#page=18",
    )
    defined_for = StateCode.MI

    adds = [
        "mi_standard_deduction_tier_two",
        "mi_standard_deduction_tier_three",
    ]



from policyengine_us.model_api import *


class de_elderly_or_disabled_income_exclusion_indiv(Variable):
    value_type = float
    entity = Person
    label = "Delaware individual aged or disabled exclusion when married filing sepaartely"
    unit = USD
    definition_period = YEAR
    reference = "https://revenuefiles.delaware.gov/2022/PIT-RES_TY22_2022-01_PaperInteractive.pdf#page=1"
    defined_for = "de_elderly_or_disabled_income_exclusion_eligible_person"

    def formula(person, period, parameters):
        # First get their filing status.
        filing_status = person.tax_unit(
            "state_filing_status_if_married_filing_separately_on_same_return",
            period,
        )
        p = parameters(
            period
        ).gov.states.de.tax.income.subtractions.exclusions.elderly_or_disabled

        return p.amount[filing_status]



from policyengine_us.model_api import *


class mi_income_tax_before_non_refundable_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "Michigan income tax before non-refundable credits"
    defined_for = StateCode.MI
    unit = USD
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        taxable_income = tax_unit("mi_taxable_income", period)
        rate = parameters(period).gov.states.mi.tax.income.rate
        return taxable_income * rate



from policyengine_us.model_api import *


class or_wfhdc_household_income(Variable):
    value_type = float
    entity = TaxUnit
    unit = USD
    label = "Household income for the Oregon working family household and dependent care credit"
    documentation = "Larger of federal and state AGI"
    definition_period = YEAR
    reference = (
        "https://www.oregon.gov/dor/forms/FormsPubs/schedule-or-wfhdc-inst_101-195-1_2022.pdf#pahe=1",
        "https://law.justia.com/codes/oregon/2021/volume-08/chapter-315/section-315-264/",
    )
    defined_for = StateCode.OR

    def formula(tax_unit, period, parameters):
        # Get household income, the larger of federal and Oregon AGI.
        federal_agi = tax_unit("adjusted_gross_income", period)
        or_agi = tax_unit("or_agi", period)
        return max_(federal_agi, or_agi)



from policyengine_us.model_api import *


class is_in_foster_care_group_home(Variable):
    value_type = bool
    entity = Person
    label = "Person is currently in a qualifying foster care group home"
    definition_period = MONTH
    default_value = False
    defined_for = "is_in_foster_care"



from policyengine_us.model_api import *


class va_child_dependent_care_expense_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Virginia child and dependent care expense deduction"
    unit = USD
    definition_period = YEAR
    reference = "https://www.tax.virginia.gov/sites/default/files/vatax-pdf/2021-760-instructions.pdf#page=29"
    defined_for = StateCode.VA

    def formula(tax_unit, period, parameters):
        expenses = tax_unit("tax_unit_childcare_expenses", period)
        cdcc_limit = tax_unit(
            "va_child_dependent_care_deduction_cdcc_limit", period
        )
        eligible_capped_expenses = min_(expenses, cdcc_limit)
        # cap further to the lowest earnings between the taxpayer and spouse
        return min_(
            eligible_capped_expenses,
            tax_unit("min_head_spouse_earned", period),
        )



from policyengine_us.model_api import *


class de_subtractions(Variable):
    value_type = float
    entity = Person
    label = "Delaware subtractions"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.DE

    adds = "gov.states.de.tax.income.subtractions.subtractions"



from policyengine_us.model_api import *


class wv_social_security_benefits_subtraction_person(Variable):
    value_type = float
    entity = Person
    label = (
        "West Virginia social security benefits subtraction for each person"
    )
    unit = USD
    definition_period = YEAR
    reference = (
        # West Virginia Personal Income Tax Forms And Instructions 2020 LINE 33
        "https://tax.wv.gov/Documents/TaxForms/2020/it140.booklet.pdf#page=24",
        # West Virginia Personal Income Tax Forms And Instructions 2020 LINE 32
        "https://tax.wv.gov/Documents/TaxForms/2021/it140.booklet.pdf#page=24",
        # West Virginia Personal Income Tax Forms And Instructions 2022 LINE 32
        "https://tax.wv.gov/Documents/PIT/2022/PersonalIncomeTaxFormsAndInstructions.2022.pdf#page=25",
        # Code of West Virginia §11-21-12 (c)(8)(A) - (c)(8)(F)
        "https://code.wvlegislature.gov/11-21-12/",
    )
    defined_for = StateCode.WV

    def formula(person, period, parameters):
        p = parameters(
            period
        ).gov.states.wv.tax.income.subtractions.social_security_benefits
        head_or_spouse = person("is_tax_unit_head_or_spouse", period)
        # Federal adjusted gross income includes
        # social security benefits paid by the Social Security Administration as Old Age,
        # Survivors and Disability Insurance Benefits
        taxable_ss = person("taxable_social_security", period)
        amount_if_eligible = taxable_ss * p.rate
        base_amount = amount_if_eligible * head_or_spouse
        eligible = person.tax_unit(
            "wv_social_security_benefits_subtraction_eligible", period
        )
        if p.social_security_benefits_above_income_limit.applies:
            multiplier = where(
                eligible, 1, p.social_security_benefits_above_income_limit.rate
            )
            return base_amount * multiplier
        return base_amount * eligible



from policyengine_us.model_api import *


class ky_itemized_deductions_indiv(Variable):
    value_type = float
    entity = Person
    label = "Kentucky itemized deductions when married couples file separately"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://revenue.ky.gov/Forms/Form%20740%20Schedule%20A%202022.pdf"
        "https://law.justia.com/codes/kentucky/2022/chapter-141/section-141-019/"  # (2)(i)
    )
    defined_for = StateCode.KY

    def formula(person, period, parameters):
        head_or_spouse = person("is_tax_unit_head_or_spouse", period)
        ky_agi = person("ky_agi", period) * head_or_spouse
        ky_agi_sum = person.tax_unit.sum(ky_agi)
        ky_agi_proportion = np.zeros_like(ky_agi_sum)
        mask = ky_agi_sum > 0
        ky_agi_proportion[mask] = ky_agi[mask] / ky_agi_sum[mask]
        itemized_deductions = person.tax_unit(
            "ky_itemized_deductions_unit", period
        )

        return ky_agi_proportion * itemized_deductions



from policyengine_us.model_api import *


class mi_expanded_retirement_benefits_deduction_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Eligible for the Michigan expanded retirement benefits deduction"
    definition_period = YEAR
    reference = (
        "http://legislature.mi.gov/doc.aspx?mcl-206-30",  # (10)
        "https://www.michigan.gov/taxes/iit/retirement-and-pension-benefits/michigan-standard-deduction",
        "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2023/2023-IIT-Forms/BOOK_MI-1040.pdf#page=20",
    )
    defined_for = StateCode.MI

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.mi.tax.income.deductions.retirement_benefits.expanded

        if p.availability:
            older_spouse_birth_year = tax_unit(
                "older_spouse_birth_year", period
            )
            return p.birth_year.calc(older_spouse_birth_year)
        else:
            return False



from policyengine_us.model_api import *


class nc_tanf_household_size(Variable):
    value_type = int
    entity = SPMUnit
    label = "NC TANF household size excludes SSI recipients"
    reference = "https://policies.ncdhhs.gov/wp-content/uploads/wf-114_1-2-2024.pdf#page=2"
    definition_period = YEAR
    defined_for = StateCode.NC

    def formula(spm_unit, period, parameters):
        ssi_income = spm_unit.members("ssi", period)

        # Eligible members are those with no SSI income
        eligible_members = ssi_income <= 0

        return eligible_members.sum()



from policyengine_us.model_api import *


class la_eitc(Variable):
    value_type = float
    entity = TaxUnit
    label = "Louisiana Earned Income Tax Credit"
    unit = USD
    definition_period = YEAR
    reference = "https://www.legis.la.gov/legis/Law.aspx?d=453085"
    defined_for = StateCode.LA

    def formula(tax_unit, period, parameters):
        federal_eitc = tax_unit("eitc", period)
        p = parameters(period).gov.states.la.tax.income.credits
        return federal_eitc * p.eitc.match



from policyengine_us.model_api import *


class ca_pre_exemption_amti(Variable):
    value_type = float
    entity = TaxUnit
    label = "California pre-exemption alternative minimum taxable income"
    defined_for = StateCode.CA
    unit = USD
    definition_period = YEAR
    reference = "https://www.ftb.ca.gov/forms/2022/2022-540-p.pdf"

    # Line 19
    adds = [
        "ca_amti_adjustments",
        "ca_taxable_income",
        "ca_itemized_deductions_pre_limitation",
    ]



from policyengine_us.model_api import *


class ssi_earned_income(Variable):
    value_type = float
    entity = Person
    label = "SSI earned income"
    unit = USD
    definition_period = YEAR

    adds = "gov.ssa.ssi.income.sources.earned"



from policyengine_us.model_api import *


class ca_tanf_vehicle_value_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = (
        "Eligible child for the California CalWORKs based on the vehicle value"
    )
    definition_period = YEAR
    defined_for = StateCode.CA
    reference = "http://epolicy.dpss.lacounty.gov/epolicy/epolicy/server/general/projects_responsive/ePolicyMaster/index.htm?&area=general&type=responsivehelp&ctxid=&project=ePolicyMaster#t=mergedProjects/CalWORKs/CalWORKs/42-215_Determining_Value_of_Property_Vehicles/42-215_Determining_Value_of_Property_Vehicles.htm"

    def formula(spm_unit, period, parameters):
        vehicle_value = add(spm_unit, period, ["household_vehicles_value"])
        p = parameters(period).gov.states.ca.cdss.tanf.cash.resources.limit
        return vehicle_value <= p.vehicle



from policyengine_us.model_api import *


class la_general_relief_housing_subsidy_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = (
        "Eligible for the Los Angeles County General Relief Housing Subsidy"
    )
    definition_period = YEAR
    # Person has to be a resident of LA County
    defined_for = "la_general_relief_eligible"
    reference = "https://dpss.lacounty.gov/en/cash/gr/housing.html"

    def formula(spm_unit, period, parameters):
        household = spm_unit.household
        # Person has to experience homelessness
        # Considering people who are about to loose housing as homeless
        homeless = household("is_homeless", period)
        # Person has to be aprt of other specific programs to be eligible for the housing subsidy
        program_eligible = spm_unit(
            "la_general_relief_housing_subsidy_program_eligible", period
        )
        # Person or couple can not have dependents
        dependents = add(spm_unit, period, ["tax_unit_dependents"])
        # General relief cannot be under the rent subsidy amounts
        gr_amount_eligible = spm_unit(
            "la_general_relief_housing_subsidy_base_amount_eligible", period
        )
        return (
            program_eligible
            & homeless
            & (dependents == 0)
            & gr_amount_eligible
        )



from policyengine_us.model_api import *


class puerto_rico_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "Income from Puerto Rico"
    unit = USD
    documentation = "Income generated in Puerto Rico by any individual who is a bona fide resident."
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/933"



from policyengine_us.model_api import *


class investment_in_529_plan(Variable):
    value_type = float
    entity = TaxUnit
    label = "529 plan investment"
    unit = USD
    documentation = "Amount invested in a 529 savings plan."
    definition_period = YEAR

    adds = ["investment_in_529_plan_indv"]



from policyengine_us.model_api import *


class ca_calworks_child_care(Variable):
    value_type = float
    entity = SPMUnit
    label = "California CalWORKs Child Care final payment"
    unit = USD
    definition_period = MONTH
    defined_for = "ca_calworks_child_care_eligible"

    adds = ["ca_calworks_child_care_payment"]



from policyengine_us.model_api import *


class spm_unit_market_income(Variable):
    value_type = float
    entity = SPMUnit
    label = "Market income"
    definition_period = YEAR
    unit = USD

    adds = ["market_income"]



from policyengine_us.model_api import *


class ia_reduced_tax(Variable):
    value_type = float
    entity = TaxUnit
    label = "Iowa income tax reduced amount for single tax units"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://tax.iowa.gov/sites/default/files/2022-01/IA1040%2841-001%29.pdf"
        "https://tax.iowa.gov/sites/default/files/2023-01/2021%20Expanded%20Instructions_010323.pdf#page=60"
        "https://tax.iowa.gov/sites/default/files/2023-01/2022IA1040%2841001%29.pdf"
        "https://tax.iowa.gov/sites/default/files/2023-03/2022%20Expanded%20Instructions_022023.pdf#page=60"
    )
    defined_for = StateCode.IA

    def formula(tax_unit, period, parameters):
        # logic follows Tax Reduction Worksheet
        modified_income = tax_unit("ia_modified_income", period)  # Line 1
        p = parameters(period).gov.states.ia.tax.income.tax_reduction
        is_elderly = tax_unit("age_head", period) >= p.elderly_age
        modified_income_threshold = where(  # Line 2
            is_elderly, p.threshold.elderly, p.threshold.nonelderly
        )
        amount = max_(0, modified_income - modified_income_threshold)  # Line 3
        # reduced tax available only to single tax units
        filing_status = tax_unit("filing_status", period)
        is_single = filing_status == filing_status.possible_values.SINGLE
        return where(is_single, amount, np.inf)
        # see ia_income_tax_before_refundable_credits formula for variable use



from policyengine_us.model_api import *


class ny_ctc_post_2024_phase_out(Variable):
    value_type = float
    entity = TaxUnit
    label = "New York CTC post-2024 phase-out amount"
    documentation = "Amount by which New York CTC is reduced due to income phase-out under post-2024 rules"
    unit = USD
    definition_period = YEAR
    reference = "https://www.nysenate.gov/legislation/laws/TAX/606"  # (c-1)
    defined_for = "ny_ctc_post_2024_eligible"

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.ny.tax.income.credits.ctc
        agi = tax_unit("adjusted_gross_income", period)
        filing_status = tax_unit("filing_status", period)
        base_credit = tax_unit("ny_ctc_post_2024_base", period)

        # Only apply phase-out if there's a base credit and phase-out rate > 0
        phase_out_threshold = p.post_2024.phase_out.threshold[filing_status]
        excess_income = max_(agi - phase_out_threshold, 0)
        # Round up to nearest increment for phase-out calculation
        increment = p.post_2024.phase_out.increment
        excess_increments = (excess_income + increment - 1) // increment
        phase_out_amount = excess_increments * p.post_2024.phase_out.rate

        # Apply phase-out only where there's a base credit and phase-out rate > 0
        has_base_credit = base_credit > 0
        has_phase_out_rate = p.post_2024.phase_out.rate > 0
        apply_phase_out = has_base_credit & has_phase_out_rate

        return apply_phase_out * phase_out_amount



from policyengine_us.model_api import *


class nm_hundred_year_exemption(Variable):
    value_type = float
    entity = TaxUnit
    label = "New Mexico hundred year exemption"
    unit = USD
    definition_period = YEAR
    reference = "https://nmonesource.com/nmos/nmsa/en/item/4340/index.do#!fragment/zoupio-_Toc140503677/BQCwhgziBcwMYgK4DsDWszIQewE4BUBTADwBdoAvbRABwEtsBaAfX2zgEYAWABgFYeAZgBsAdlEBKADTJspQhACKiQrgCe0AOSapEQmFwJlqjdt37DIAMp5SAIQ0AlAKIAZZwDUAggDkAws5SpGAARtCk7BISQA"
    defined_for = StateCode.NM

    def formula(tax_unit, period, parameters):
        age_head = tax_unit("age_head", period)
        age_spouse = tax_unit("age_spouse", period)
        p = parameters(period).gov.states.nm.tax.income.exemptions.hundred_year
        head_eligible = age_head >= p.age_eligibility
        spouse_eligible = age_spouse >= p.age_eligibility
        filing_status = tax_unit("filing_status", period)
        joint = filing_status == filing_status.possible_values.JOINT
        # Halve the exemption if only one of head and spouse is eligible of a joint filer.
        denominator = where(joint, 2, 1)
        numerator = head_eligible.astype(int) + spouse_eligible.astype(int)
        # Exempt AGI apportioned among eligible spouses.
        # That is, assume all income is community property.
        return (
            tax_unit("adjusted_gross_income", period) * numerator / denominator
        )



from policyengine_us.model_api import *


class count_529_contribution_beneficiaries(Variable):
    value_type = int
    entity = Person
    label = "Number of beneficiaries to 529 college savings plan contributions"
    definition_period = YEAR



from policyengine_us.model_api import *


class ar_inflation_relief_credit_person(Variable):
    value_type = float
    entity = Person
    label = "Arkansas inflation relief income-tax credit for each individual"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.AR

    def formula(person, period, parameters):
        # the Arkansas inflationary relief tax credit is calculated based on net taxable income
        # mentioned in individual income tax return form AR1000F line 28
        filing_separately = person.tax_unit("ar_files_separately", period)
        income_joint = person("ar_taxable_income_joint", period)
        income_indiv = person("ar_taxable_income_indiv", period)
        # When filing separartely, the credit is calculated based on individual income
        income = where(
            filing_separately, income_indiv, person.tax_unit.sum(income_joint)
        )
        p = parameters(
            period
        ).gov.states.ar.tax.income.credits.inflationary_relief
        filing_status = person.tax_unit("filing_status", period)
        max_amount = p.max_amount[filing_status]
        reduction_start = p.reduction.start[filing_status]
        increment = p.reduction.increment[filing_status]
        reduction_per_increment = p.reduction.amount[filing_status]
        excess = max_(income - reduction_start, 0)
        increments = np.ceil(excess / increment)
        total_reduction_amount = increments * reduction_per_increment
        # Attribute the maximum amount to each spouse equally when married filing jointly
        joint = filing_status == filing_status.possible_values.JOINT
        divisor = where(joint, 2, 1)
        return max_(max_amount - total_reduction_amount, 0) / divisor



from policyengine_us.model_api import *


class ma_tafdc_payment_standard(Variable):
    value_type = float
    unit = USD
    entity = SPMUnit
    label = "Massachusetts Temporary Assistance for Families with Dependent Children (TAFDC) payment standard"
    definition_period = MONTH
    reference = (
        "https://www.law.cornell.edu/regulations/massachusetts/106-CMR-704-420"
    )
    defined_for = StateCode.MA

    def formula(spm_unit, period, parameters):
        in_public_housing = spm_unit.household("is_in_public_housing", period)
        p = parameters(
            period
        ).gov.states.ma.dta.tcap.tafdc.eligibility.income_limit
        unit_size = spm_unit("spm_unit_size", period)
        capped_unit_size = min_(unit_size, p.max_unit_size)
        additional_people_in_unit = max_(0, unit_size - p.max_unit_size)
        # Calculate the base income limit for non-teen parents
        ps_non_teen_parent = where(
            in_public_housing,
            p.base.public_housing.calc(capped_unit_size),
            p.base.private_housing.calc(capped_unit_size),
        )
        additional_person_ps = p.additional_person * additional_people_in_unit
        return ps_non_teen_parent + additional_person_ps



from policyengine_us.model_api import *


class ri_eitc(Variable):
    value_type = float
    entity = TaxUnit
    label = "Rhode Island earned income tax credit"
    defined_for = StateCode.RI
    unit = USD
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        federal_eitc = tax_unit("eitc", period)
        rate = parameters(period).gov.states.ri.tax.income.credits.eitc.match
        return federal_eitc * rate



from policyengine_us.model_api import *


class living_arrangements_allow_for_food_preparation(Variable):
    value_type = bool
    entity = Household
    label = "Living arrangements allow for food preparation"
    definition_period = YEAR



from policyengine_us.model_api import *


class pa_tanf_resources_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Meets Pennsylvania TANF resource limit"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.PA

    def formula(spm_unit, period, parameters):
        p = parameters(period).gov.states.pa.dhs.tanf
        resources = spm_unit("pa_tanf_countable_resources", period)
        return resources <= p.resource_limit



from policyengine_us.model_api import *


class mt_files_separately(Variable):
    value_type = bool
    entity = TaxUnit
    label = "married couple files separately on Montana tax return"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MT

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.mt.tax.income
        if p.married_filing_separately_on_same_return_allowed:
            itax_indiv = add(tax_unit, period, ["mt_income_tax_indiv"])
            itax_joint = tax_unit("mt_income_tax_joint", period)
            return itax_indiv < itax_joint
        return False



from policyengine_us.model_api import *


class amt_tax_including_cg(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Alternative Minimum Tax computed using the capital gains rates"
    unit = USD
    documentation = "Alternative Minimum Tax (AMT) liability computed using the capital gains rates, Form 6251, Part III"
    reference = "https://www.irs.gov/pub/irs-pdf/f6251.pdf"

    def formula(tax_unit, period, parameters):
        # Line 12
        reduced_income = tax_unit("amt_income_less_exemptions", period)
        # Line 13 - schedule D line 13
        cg_distributions = tax_unit("dwks13", period)
        # Line 14 - schedule D line 19
        section_1250_gain_worksheet = tax_unit(
            "unrecaptured_section_1250_gain", period
        )
        # Line 15 - smaller of the sum of Line 13 and Line 14 or Schedule D line 10
        total_transactions_reported = tax_unit("dwks10", period)
        capped_capital_gains = min_(
            cg_distributions + section_1250_gain_worksheet,
            total_transactions_reported,
        )
        # Line 16 - smaller of Line 15 or Line 12
        capped_income = min_(capped_capital_gains, reduced_income)
        # Line 17 - Line 12 minus Line 16
        excess_income = max_(0, reduced_income - capped_income)
        # Line 18 - Apply CG tax rates to Line 17
        p = parameters(period).gov.irs
        income_taxes_at_amt_rates = p.income.amt.brackets.calc(excess_income)
        # Line 19 - First CG tax bracket threshold
        filing_status = tax_unit("filing_status", period)
        cg_bracket = p.capital_gains.brackets.thresholds["1"][filing_status]
        # Line 20 - Schedule D Line 14
        lt_capital_loss_carryover = tax_unit("dwks14", period)
        # Line 21 - Line 20 minus Line 19
        reduced_cg_bracket = max_(0, cg_bracket - lt_capital_loss_carryover)
        # Line 22 - smaller of Line 12 or Line 13
        smaller_of_income_or_cg = min_(reduced_income, cg_distributions)
        # Line 23 - smaller of Line 22 or Line 21 (amount is taxed at 0%)
        cg_first_rate = p.capital_gains.brackets.rates["1"]
        disregarded_gains = (
            min_(smaller_of_income_or_cg, reduced_cg_bracket) * cg_first_rate
        )
        # Line 24 - Line 22 minus Line 23
        taxable_income_including_cg = max_(
            smaller_of_income_or_cg - disregarded_gains, 0
        )
        # Line 25 - Second CG tax bracket threshold
        second_cg_bracket = p.capital_gains.brackets.thresholds["2"][
            filing_status
        ]
        # Line 26 - same as line 21
        # Line 27 - Schedule D Line 21
        loss_limited_net_capital_gains = tax_unit(
            "loss_limited_net_capital_gains", period
        )
        # Line 28 - Line 26 plus Line 27
        first_cg_bracket_increased_by_loss = (
            loss_limited_net_capital_gains + reduced_cg_bracket
        )
        # Line 29 Line 25 minus Line 28
        reduced_second_cg_bracket = max_(
            0, second_cg_bracket - first_cg_bracket_increased_by_loss
        )
        # Line 30 - smaller of Line 24 or Line 29
        capped_income_including_cg = min_(
            taxable_income_including_cg, reduced_second_cg_bracket
        )
        # Line 31 - multiply Line 30by second CG tax rate
        cg_second_bracket_tax = (
            capped_income_including_cg * p.capital_gains.brackets.rates["2"]
        )
        # Line 32 - Line 23 plus Line 30
        taxed_gains = disregarded_gains + capped_income_including_cg
        # Line 33 - Line 22 minus Line 32
        excess_taxed_gains = max_(0, smaller_of_income_or_cg - taxed_gains)
        # Line 34 - multiply Line 33 by third CG tax rate
        cg_third_bracket_tax = (
            excess_taxed_gains * p.capital_gains.brackets.rates["3"]
        )
        # Line 35 - sum of Line 17, Line 32, Line 33
        final_taxed_income = excess_income + taxed_gains + excess_taxed_gains
        # Line 36 Line 12 minus Line 35
        final_excess = max_(0, reduced_income - final_taxed_income)
        # Line 37 - Multiply Line 36 by the AMT specific capital gain excess tax rate
        # The increased tax does not applied if no recaptured section 1250 gains
        unrecaptured_section_1250_gain = tax_unit(
            "unrecaptured_section_1250_gain", period
        )
        excess_tax = where(
            unrecaptured_section_1250_gain == 0,
            0,
            final_excess
            * p.income.amt.capital_gains.capital_gain_excess_tax_rate,
        )
        # Line 38 - sum of Line 18, Line 31, Line 34, and Line 37
        return (
            income_taxes_at_amt_rates
            + cg_second_bracket_tax
            + cg_third_bracket_tax
            + excess_tax
        )



from policyengine_us.model_api import *


class ca_riv_share_electricity_emergency_payment(Variable):
    value_type = float
    entity = SPMUnit
    label = "Riverside County Sharing Households Assist Riverside's Energy program (SHARE) electric emergency payment"
    unit = USD
    definition_period = YEAR
    defined_for = "ca_riv_share_eligible"
    reference = "https://riversideca.gov/utilities/residents/assistance-programs/share-english"

    def formula(spm_unit, period, parameters):
        # This electric emergency payment only assign once per 12-month period
        p = parameters(period).gov.local.ca.riv.cap.share.payment
        is_urgent = spm_unit(
            "ca_riv_share_eligible_for_emergency_payment", period
        )
        return is_urgent * p.electricity_emergency



from policyengine_us.model_api import *


class wi_married_couple_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Wisconsin married couple credit"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.revenue.wi.gov/TaxForms2021/2021-Form1f.pdf#page=4"
        "https://www.revenue.wi.gov/TaxForms2021/2021-Form1-Inst.pdf#page=21"
        "https://www.revenue.wi.gov/TaxForms2022/2022-Form1f.pdf#page=4"
        "https://www.revenue.wi.gov/TaxForms2022/2022-Form1-Inst.pdf#page=21"
        "https://docs.legis.wisconsin.gov/misc/lfb/informational_papers/january_2023/0002_individual_income_tax_informational_paper_2.pdf"
    )
    defined_for = StateCode.WI

    def formula(tax_unit, period, parameters):
        fstatus = tax_unit("filing_status", period)
        eligible = fstatus == fstatus.possible_values.JOINT
        p = parameters(period).gov.states.wi.tax.income.credits
        income = add(tax_unit.members, period, p.married_couple.income_sources)
        is_head = tax_unit.members("is_tax_unit_head", period)
        is_spouse = tax_unit.members("is_tax_unit_spouse", period)
        head_income = tax_unit.sum(is_head * income)
        spouse_income = tax_unit.sum(is_spouse * income)
        lower_income = min_(head_income, spouse_income)
        return min_(
            eligible * lower_income * p.married_couple.rate,
            p.married_couple.max,
        )



from policyengine_us.model_api import *


class net_investment_income_tax(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Net Investment Income Tax"
    reference = "https://www.law.cornell.edu/uscode/text/26/1411"
    unit = USD

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.irs.investment.net_investment_income_tax
        threshold = p.threshold[tax_unit("filing_status", period)]
        excess_agi = max_(
            0, tax_unit("adjusted_gross_income", period) - threshold
        )
        base = min_(
            max_(0, tax_unit("net_investment_income", period)),
            excess_agi,
        )
        return p.rate * base



from policyengine_us.model_api import *


class nj_property_tax_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "New Jersey property tax credit"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.nj.gov/treasury/taxation/pdf/other_forms/tgi-ee/2021/1040i.pdf#page=49"
        "https://www.nj.gov/treasury/taxation/pdf/other_forms/tgi-ee/2022/1040i.pdf#page=50"
        "https://www.state.nj.us/treasury/taxation/pdf/current/1040i.pdf#page=52"
    )
    defined_for = "nj_property_tax_credit_eligible"

    def formula(tax_unit, period, parameters):
        # calculate credit amount
        filing_status = tax_unit("filing_status", period)
        status = filing_status.possible_values
        separate = filing_status == status.SEPARATE
        cohabitating = tax_unit("cohabitating_spouses", period)
        p = parameters(period).gov.states.nj.tax.income.credits.property_tax
        credit_amount = p.amount / (1 + separate * cohabitating)

        # taking credit if not taking deduction
        taking_deduction = tax_unit("nj_taking_property_tax_deduction", period)
        return credit_amount * ~taking_deduction



from policyengine_us.model_api import *


class va_income_tax_before_non_refundable_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "Virginia income tax before non-refundable credits"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://law.lis.virginia.gov/vacodefull/title58.1/chapter3/article2/"
    )
    defined_for = "va_must_file"

    def formula(tax_unit, period, parameters):
        taxable_income = tax_unit("va_taxable_income", period)
        p = parameters(period).gov.states.va.tax.income.rates
        return p.calc(taxable_income)



from policyengine_us.model_api import *


class mt_property_tax_rebate(Variable):
    value_type = float
    entity = TaxUnit
    label = "Montana property tax rebate"
    unit = USD
    definition_period = YEAR
    reference = "https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/12/Form_2_2023_Instructions.pdf#page=5"
    defined_for = StateCode.MT

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.mt.tax.income.credits.rebate.property
        person = tax_unit.members
        mt_property_tax = person("real_estate_taxes", period)
        tax_unit_mt_property_tax = tax_unit.sum(mt_property_tax)
        return min_(p.amount, tax_unit_mt_property_tax)



from policyengine_us.model_api import *


class ma_tafdc_full_earned_income_disregard_eligible(Variable):
    value_type = bool
    entity = Person
    label = "Is eligible for the full earned income disregard under the Massachusetts Temporary Assistance for Families with Dependent Children (TAFDC)"
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/regulations/massachusetts/106-CMR-704-281"  # (A)
    defined_for = StateCode.MA

    def formula(person, period, parameters):
        p = parameters(
            period
        ).gov.states.ma.dta.tcap.tafdc.earned_income_disregard.full_disregard
        gross_income = person.spm_unit(
            "ma_tafdc_applicable_income_for_financial_eligibility", period
        )

        fpg = person.spm_unit("spm_unit_fpg", period)
        income_limit = fpg * p.fpg_limit
        return gross_income < income_limit



from policyengine_us.model_api import *


class ri_income_tax(Variable):
    value_type = float
    entity = TaxUnit
    label = "Rhode Island income tax"
    defined_for = StateCode.RI
    unit = USD
    definition_period = YEAR
    adds = ["ri_income_tax_before_refundable_credits"]
    subtracts = ["ri_refundable_credits"]



from policyengine_us.model_api import *


class ca_riv_share_eligible_for_emergency_payment(Variable):
    value_type = bool
    entity = SPMUnit
    label = "SPM Unit has urgent notice and/or disconnection notice under Riverside County SHARE program"
    definition_period = MONTH
    defined_for = "in_riv"
    reference = "https://riversideca.gov/utilities/residents/assistance-programs/share-english"



INPUT_VARIABLES = [
    "tanf_reported",
    "spm_unit_payroll_tax_reported",
    "veterans_benefits",
    "other_credits",
    "tax_exempt_ira_distributions",
    "ssi_reported",
    "tax_exempt_403b_distributions",
    "taxable_ira_distributions",
    "w2_wages_from_qualified_business",
    "roth_ira_contributions",
    "social_security_retirement",
    "domestic_production_ald",
    "social_security_disability",
    "health_savings_account_ald",
    "investment_income_elected_form_4952",
    "medical_out_of_pocket_expenses",
    "unemployment_compensation",
    "self_employment_income",
    "interest_deduction",
    "energy_efficient_home_improvement_credit",
    "spm_unit_net_income_reported",
    "roth_401k_contributions",
    "spm_unit_federal_tax_reported",
    "spm_unit_pre_subsidy_childcare_expenses",
    "taxable_pension_income",
    "keogh_distributions",
    "social_security_survivors",
    "amt_foreign_tax_credit",
    "cdcc_relevant_expenses",
    "social_security",
    "workers_compensation",
    "self_employed_health_insurance_ald",
    "state_and_local_sales_or_income_tax",
    "child_support_received",
    "child_support_expense",
    "tax_exempt_sep_distributions",
    "salt_refund_income",
    "spm_unit_total_income_reported",
    "alimony_income",
    "prior_year_minimum_tax_credit",
    "self_employed_pension_contributions",
    "taxable_unemployment_compensation",
    "employment_income_last_year",
    "taxable_sep_distributions",
    "spm_unit_capped_housing_subsidy_reported",
    "miscellaneous_income",
    "short_term_capital_gains",
    "qualified_dividend_income",
    "tax_exempt_private_pension_income",
    "rental_income",
    "student_loan_interest",
    "tax_exempt_interest_income",
    "general_business_credit",
    "early_withdrawal_penalty",
    "spm_unit_weight",
    "casualty_loss",
    "partnership_s_corp_income",
    "taxable_interest_income",
    "spm_unit_spm_threshold",
    "non_sch_d_capital_gains",
    "spm_unit_state_tax_reported",
    "spm_unit_capped_work_childcare_expenses",
    "farm_income",
    "taxable_403b_distributions",
    "qualified_tuition_expenses",
    "disability_benefits",
    "taxable_401k_distributions",
    "farm_rent_income",
    "unreported_payroll_tax",
    "traditional_ira_contributions",
    "tax_exempt_401k_distributions",
    "alimony_expense",
    "taxable_private_pension_income",
    "charitable_cash_donations",
    "spm_unit_broadband_subsidy_reported",
    "free_school_meals_reported",
    "non_qualified_dividend_income",
    "excess_withheld_payroll_tax",
    "health_insurance_premiums",
    "spm_unit_energy_subsidy_reported",
    "real_estate_taxes",
    "estate_income",
    "misc_deduction",
    "self_employed_pension_contribution_ald",
    "spm_unit_wic_reported",
    "unrecaptured_section_1250_gain",
    "snap_reported",
    "long_term_capital_gains_on_collectibles",
    "self_employment_income_last_year",
    "tax_exempt_pension_income",
    "employment_income",
    "foreign_tax_credit",
    "pre_tax_contributions",
    "social_security_dependents",
    "american_opportunity_credit",
    "charitable_non_cash_donations",
    "recapture_of_investment_credit",
    "savers_credit",
    "educator_expense",
    "long_term_capital_gains",
    "traditional_401k_contributions",
    "strike_benefits",
    "other_medical_expenses",
    "over_the_counter_health_expenses",
    "medicare_part_b_premiums",
    "health_insurance_premiums_without_medicare_part_b",
]


def add_default_uprating(system):
    for variable in system.variables.values():
        if (variable.name in INPUT_VARIABLES) and (variable.uprating is None):
            variable.uprating = (
                "calibration.gov.cbo.income_by_source.adjusted_gross_income"
            )
        system.variables[variable.name] = variable



from policyengine_us.model_api import *


class il_dependent_exemption(Variable):
    value_type = float
    entity = TaxUnit
    label = "Illinois dependent exemption"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.IL

    def formula(tax_unit, period, parameters):
        amount = parameters(
            period
        ).gov.states.il.tax.income.exemption.dependent

        return amount * tax_unit("tax_unit_dependents", period)



from policyengine_us.model_api import *


class ks_income_tax_before_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "Kansas income tax before credits"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.ksrevenue.gov/pdf/ip21.pdf"
        "https://www.ksrevenue.gov/pdf/ip22.pdf"
    )
    defined_for = StateCode.KS

    def formula(tax_unit, period, parameters):
        filing_status = tax_unit("filing_status", period)
        taxable_income = tax_unit("ks_taxable_income", period)
        p = parameters(period).gov.states.ks.tax.income.rates
        tax = where(
            filing_status == filing_status.possible_values.JOINT,
            p.joint.calc(taxable_income),
            p.other.calc(taxable_income),
        )
        zero_tax_agi_threshold = p.zero_tax_threshold[filing_status]
        agi = tax_unit("ks_agi", period)
        return where(agi <= zero_tax_agi_threshold, 0, tax)



from policyengine_us.model_api import *


class de_agi_joint(Variable):
    value_type = float
    entity = Person
    label = "Delaware adjusted gross income for each individual whe married filing jointly"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.DE

    def formula(person, period, parameters):
        pre_exclusions_agi = person("de_pre_exclusions_agi", period)
        indv_exclusions = person(
            "de_elderly_or_disabled_income_exclusion_joint", period
        )
        net_income = max_(pre_exclusions_agi - indv_exclusions, 0)
        # allocate any dependent gross income to tax unit head
        is_head = person("is_tax_unit_head", period)
        return person.tax_unit.sum(net_income) * is_head



from policyengine_us.model_api import *


class childcare_hours_per_day(Variable):
    value_type = float
    entity = Person
    label = "Child care hours per day"
    definition_period = YEAR
    unit = "hour"



from policyengine_us.model_api import *


class retired_from_federal_government(Variable):
    value_type = bool
    entity = Person
    label = "Retired from Federal Government"
    definition_period = YEAR
    defined_for = "is_retired"
    default_value = False



from policyengine_us.model_api import *


class medicaid_work_requirement_eligible(Variable):
    value_type = bool
    entity = Person
    label = "Eligible person for Medicaid via work requirement"
    definition_period = YEAR
    reference = (
        "https://www.congress.gov/bill/119th-congress/house-bill/1/text"
    )

    def formula(person, period, parameters):
        p = parameters(period).gov.hhs.medicaid.eligibility.work_requirements
        # Works no less than 80 hours p.680 (2)(A)
        monthly_hours_worked = person("monthly_hours_worked", period)
        meets_monthly_work_hours = (
            monthly_hours_worked >= p.monthly_hours_threshold
        )
        # The individual is enrolled in an educational program at least half-time. p.680 (2)(D)
        is_full_time_student = person("is_full_time_student", period)
        # pregnant or postpartum medical assistance p.681 (3)(A)(i)(II)(bb)
        is_pregnant = person("is_pregnant", period)
        # Has attained age of 19 and is under 65 is require to work p.693 (bb)
        age = person("age", period)
        work_required_age = p.age_range.calc(age)
        # parent, guardian, caretaker of a disabled person
        is_dependent = person("is_tax_unit_dependent", period)
        is_disabled = person("is_disabled", period)
        has_disabled = person.tax_unit.any(is_dependent & is_disabled)
        # veteran and is_permanently_and_totally_disabled p.694 (IV)
        is_veteran = person("is_veteran", period)
        is_permanently_and_totally_disabled = person(
            "is_permanently_and_totally_disabled", period
        )
        eligible_veteran = is_veteran & is_permanently_and_totally_disabled
        # blind or disabled or is_incapable_of_self_care p.694 (V)
        is_blind = person("is_blind", period)
        is_incapable_of_self_care = person("is_incapable_of_self_care", period)
        eligible_disabled = is_blind | is_disabled | is_incapable_of_self_care
        # parent, guardian, caretaker of a dependent child 13 years of age or under  p.694 (III)
        child_age_eligible = age <= p.dependent_age_limit
        has_eligible_dependent_child = person.tax_unit.any(
            is_dependent & child_age_eligible
        )
        exempted_from_work = (
            is_full_time_student
            | is_pregnant
            | has_disabled
            | eligible_veteran
            | eligible_disabled
        )
        meets_base_requirement = meets_monthly_work_hours | exempted_from_work
        meets_conditions = (
            meets_base_requirement | has_eligible_dependent_child
        )
        return where(work_required_age, meets_conditions, True)



from policyengine_us.model_api import *


class nm_net_capital_gains_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "New Mexico net capital gain deduction"
    unit = USD
    definition_period = YEAR
    reference = "https://nmonesource.com/nmos/nmsa/en/item/4340/index.do#!fragment/zoupio-_Toc140503882/BQCwhgziBcwMYgK4DsDWszIQewE4BUBTADwBdoAvbRABwEtsBaAfX2zgEYAWABgFYeAZgAcwgEwBKADTJspQhACKiQrgCe0AOSapEQmFwJlqjdt37DIAMp5SAIQ0AlAKIAZZwDUAggDkAws5SpGAARtCk7BISQA"
    defined_for = StateCode.NM

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.nm.tax.income.deductions.net_capital_gains
        net_capital_gains = max_(0, add(tax_unit, period, ["capital_gains"]))
        # Filers can deduct 100% of CG up to a cap, or 40% uncapped, whichever is greater.
        uncapped_element = p.uncapped_element_percent * net_capital_gains
        filing_status = tax_unit("filing_status", period)
        separate = filing_status == filing_status.possible_values.SEPARATE
        # Halve the deduction if filing separately.
        denominator = where(separate, 2, 1)
        capped_element = p.capped_element.calc(net_capital_gains)
        numerator = max_(uncapped_element, capped_element)
        return numerator / denominator



from policyengine_us.model_api import *


class pr_refundable_ctc_social_security_tax(Variable):
    value_type = float
    entity = TaxUnit
    label = "Puerto Rico social security and medicare taxes for refundable Child Tax Credit"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.irs.gov/pub/irs-pdf/f1040s8.pdf"
        "https://www.law.cornell.edu/uscode/text/26/24#h_4_A"
    )

    # line 23
    adds = "gov.irs.credits.ctc.refundable.social_security.add"



from policyengine_us.model_api import *


class pell_grant_head_contribution(Variable):
    value_type = float
    entity = Person
    unit = USD
    label = "Pell Grant head contribution"
    definition_period = YEAR

    def formula(person, period, parameters):
        dependents = person.tax_unit(
            "pell_grant_dependents_in_college", period
        )
        available_income = person("pell_grant_head_available_income", period)
        assets = person("pell_grant_contribution_from_assets", period)
        adjusted_available_income = available_income + assets
        formula = person("pell_grant_formula", period)
        uses_efc = person("pell_grant_uses_efc", period)
        uses_sai = person("pell_grant_uses_sai", period)
        p = parameters(period).gov.ed.pell_grant.head
        positive_head_contribution = p.marginal_rate.calc(
            adjusted_available_income
        )
        negative_head_contribution = max_(
            adjusted_available_income * p.negative_rate, p.min_contribution
        )
        total_head_contribution = where(
            adjusted_available_income >= 0,
            positive_head_contribution,
            negative_head_contribution,
        )
        total = where(
            formula == formula.possible_values.B,
            adjusted_available_income,
            total_head_contribution,
        )
        # Return amount per dependent, using a mask to avoid division by zero.
        amount_per_dependent = np.zeros_like(total)
        mask = dependents > 0
        amount_per_dependent[mask] = total[mask] / dependents[mask]

        return select([uses_efc, uses_sai], [amount_per_dependent, total])



from policyengine_us.model_api import *


class oh_non_public_school_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "Ohio Nonchartered, Nonpublic, School Tuition Credit AGI Credit"
    unit = USD
    definition_period = YEAR
    reference = "https://tax.ohio.gov/static/forms/ohio_individual/individual/2021/pit-it1040-booklet.pdf#page=21"
    defined_for = StateCode.OH

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.oh.tax.income.credits
        agi = tax_unit("adjusted_gross_income", period)
        person = tax_unit.members
        tuition = tax_unit.sum(person("non_public_school_tuition", period))
        cap = p.non_public_tuition.calc(agi)
        return min_(tuition, cap)



from policyengine_us.model_api import *


class il_income_tax_before_non_refundable_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "IL income tax before credits"
    unit = USD
    definition_period = YEAR

    defined_for = StateCode.IL

    def formula(tax_unit, period, parameters):
        recapture_of_investment_credit = tax_unit(
            "recapture_of_investment_credit", period
        )
        rate = parameters(period).gov.states.il.tax.income.rate
        return (
            tax_unit("il_taxable_income", period) * rate
            + recapture_of_investment_credit
        )



from policyengine_us.model_api import *


class ca_state_supplement_blind_amount(Variable):
    value_type = float
    entity = SPMUnit
    label = "California SSI state supplement blind amount"
    unit = USD
    definition_period = MONTH
    defined_for = StateCode.CA
    reference = "https://leginfo.legislature.ca.gov/faces/codes_displaySection.xhtml?lawCode=WIC&sectionNum=12200"

    def formula(spm_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.ca.cdss.state_supplement.payment_standard
        person = spm_unit.members
        head_or_spouse = person("is_tax_unit_head_or_spouse", period)
        # Blind amount
        blind = person("is_blind", period) * head_or_spouse
        eligible = person("ca_state_supplement_eligible_person", period)
        blind_count = spm_unit.sum(blind * eligible)
        is_married = spm_unit("spm_unit_is_married", period)
        blind_married_amount = select(
            [blind_count >= 2, blind_count == 1],
            [p.blind.married.two_blind, p.blind.married.one_blind],
            default=0,
        )
        total_amount_if_eligible = where(
            is_married, blind_married_amount, p.blind.single
        )
        return where(blind_count == 0, 0, total_amount_if_eligible)



from policyengine_us.model_api import *


class is_aca_eshi_eligible(Variable):
    value_type = bool
    entity = Person
    label = (
        "Person is eligible for employer-sponsored health insurance "
        "under ACA rules"  # if this is true, person is ineligible for ACA PTC
    )
    definition_period = YEAR

    def formula(person, period, parameters):

        has = person("has_esi", period)  # has ESI
        offered = person(
            "offered_aca_disqualifying_esi", period
        )  # being offered disqualifying ESI
        return has | offered



from policyengine_us.model_api import *


class estate_tax(Variable):
    value_type = float
    entity = Person
    label = "Estate tax"
    unit = USD
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/2001#b_1"

    def formula(person, period, parameters):
        estate_tax_before_credits = person("estate_tax_before_credits", period)
        estate_tax_credit = person("estate_tax_credit", period)
        return max_(0, estate_tax_before_credits - estate_tax_credit)



from policyengine_us.model_api import *


class id_grocery_credit_base(Variable):
    value_type = float
    entity = Person
    label = "Idaho base grocery credit"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.ID
    reference = (
        "https://law.justia.com/codes/idaho/2022/title-63/chapter-30/section-63-3024a/",
        "https://tax.idaho.gov/wp-content/uploads/forms/EFO00089/EFO00089_12-30-2022.pdf#page=7",
    )

    def formula(person, period, parameters):
        return parameters(
            period
        ).gov.states.id.tax.income.credits.grocery.base.amount



from policyengine_us.model_api import *


class va_agi(Variable):
    value_type = float
    entity = TaxUnit
    label = "Virginia adjusted federal adjusted gross income"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.tax.virginia.gov/sites/default/files/taxforms/individual-income-tax/2022/760-2022.pdf",
        "https://law.lis.virginia.gov/vacodefull/title58.1/chapter3/article2/",
    )
    defined_for = StateCode.VA

    adds = ["adjusted_gross_income", "va_additions"]

    subtracts = ["va_subtractions"]



from policyengine_us.model_api import *


class is_wic_at_nutritional_risk(Variable):
    value_type = bool
    entity = Person
    definition_period = MONTH
    documentation = "Assessed as being at nutritional risk for the Special Supplemental Nutrition Program for Women, Infants and Children (WIC)"
    label = "At nutritional risk for WIC"
    reference = "https://www.law.cornell.edu/uscode/text/42/1786#b_8"

    def formula(person, period, parameters):
        # Assign households WIC nutritional risk status probabilistically in microsimulation.
        # Assume all meet qualification in individual simulation.
        if person.simulation.dataset is not None:
            wic_reported = person("receives_wic", period)
            meets_income_test = person.spm_unit(
                "meets_wic_income_test", period
            )
            meets_categorical_test = person(
                "meets_wic_categorical_eligibility", period
            )
            category = person("wic_category", period)
            risk = parameters(period).gov.usda.wic.nutritional_risk
            imputed_risk = (random(person) < risk[category]) & (
                meets_income_test | meets_categorical_test
            )
            return wic_reported | imputed_risk
        else:
            return True



from policyengine_us.model_api import *


class il_aabd_countable_vehicle_value(Variable):
    value_type = float
    entity = SPMUnit
    definition_period = MONTH
    label = "Illinois Aid to the Aged, Blind or Disabled (AABD) countable vehicles value"
    reference = (
        "https://www.law.cornell.edu/regulations/illinois/Ill-Admin-Code-tit-89-SS-113.141",
    )
    defined_for = StateCode.IL

    def formula(spm_unit, period, parameters):
        p = parameters(period).gov.states.il.dhs.aabd.asset.vehicle_exemption
        vehicle_count = spm_unit.household("household_vehicles_owned", period)
        total_vehicle_value = (
            spm_unit.household("household_vehicles_value", period)
            * MONTHS_IN_YEAR
        )
        avg_vehicle_value = np.zeros_like(vehicle_count)
        mask = vehicle_count != 0
        avg_vehicle_value[mask] = (
            total_vehicle_value[mask] / vehicle_count[mask]
        )
        vehicle_is_essential = spm_unit("il_aabd_vehicle_is_essential", period)
        # Household can exclude one vehicle completely regardless of value if it has an essential vehicle.
        full_exemption = avg_vehicle_value
        # If household does not meet essential vehicle requirements, exempt one vehicle up to $4500 in household
        partial_exemption = min_(avg_vehicle_value, p.amount)
        vehicle_exemption = where(
            vehicle_is_essential, full_exemption, partial_exemption
        )
        return max_(total_vehicle_value - vehicle_exemption, 0)



from policyengine_us.model_api import *


class oh_personal_exemptions_eligible_person(Variable):
    value_type = bool
    entity = Person
    label = "Eligible person for the Ohio Exemption Credit"
    definition_period = YEAR
    reference = (
        "https://tax.ohio.gov/static/forms/ohio_individual/individual/2021/pit-it1040-booklet.pdf#page=14",
    )
    defined_for = StateCode.OH

    def formula(person, period, parameters):
        # The personal exemption is provided for the head and spouse
        # if they are not claimed as a dependent elsewhere
        head_or_spouse = person("is_tax_unit_head_or_spouse", period)
        dependent_on_another_return = person(
            "claimed_as_dependent_on_another_return", period
        )
        # The personal exemption is also provided to dependents
        dependent = person("is_tax_unit_dependent", period)
        return (~dependent_on_another_return & head_or_spouse) | dependent



from policyengine_us.model_api import *


class charitable_deduction_for_non_itemizers(Variable):
    value_type = float
    entity = TaxUnit
    label = "Charitable deduction for non-itemizers"
    unit = USD
    documentation = "Charitable deduction amount for non-itemizers."
    definition_period = YEAR
    reference = "https://www.irs.gov/newsroom/expanded-tax-benefits-help-individuals-and-businesses-give-to-charity-during-2021-deductions-up-to-600-available-for-cash-donations-by-non-itemizers"

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.irs.deductions.itemized.charity
        cash_donations = add(tax_unit, period, ["charitable_cash_donations"])
        filing_status = tax_unit("filing_status", period)
        return min_(p.non_itemizers_amount[filing_status], cash_donations)



from policyengine_us.model_api import *


class nm_social_security_income_exemption(Variable):
    value_type = float
    entity = TaxUnit
    label = "New Mexico social security income exemption"
    defined_for = StateCode.NM
    unit = USD
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        # This is the amount of Social Security included in AGI.
        social_security_income = tax_unit(
            "tax_unit_taxable_social_security", period
        )
        filing_status = tax_unit("filing_status", period)
        p = parameters(
            period
        ).gov.states.nm.tax.income.exemptions.social_security_income
        agi = tax_unit("adjusted_gross_income", period)
        income_limit = agi <= p.income_limit[filing_status]
        return where(income_limit, social_security_income, 0)



from policyengine_us.model_api import *


class self_employed_health_insurance_ald_person(Variable):
    value_type = float
    entity = Person
    label = "Self-employed health insurance ALD for each person"
    unit = USD
    documentation = "Personal above-the-line deduction for self-employed health insurance contributions."
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/162#l"

    def formula(person, period, parameters):
        earnings = max_(0, person("self_employment_income", period))
        premiums = person("self_employed_health_insurance_premiums", period)
        return min_(earnings, premiums)



from policyengine_us.model_api import *


class energy_efficient_central_air_conditioner_expenditures(Variable):
    value_type = float
    entity = TaxUnit
    label = "Expenditures on energy efficient central air conditioners"
    # Before the Inflation Reduction Act, these had to be the highest
    # efficiency tier established by the Consortium for Energy Efficiency as
    # of 2009.
    # After the Inflation Reduction Act, they had to be the highest efficiency
    # tier (not including any advanced tier) established by the Consortium for
    # Energy Efficiency as of the beginning of the calendar year in which the
    # property is placed in service.
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.law.cornell.edu/uscode/text/26/25C#d_3_C",
        "https://www.democrats.senate.gov/imo/media/doc/inflation_reduction_act_of_2022.pdf#page=342",
    )



from policyengine_us.model_api import *


class in_homeowners_property_tax_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Indiana homeowner's residential property tax deduction"
    unit = USD
    definition_period = YEAR
    reference = "http://iga.in.gov/legislative/laws/2021/ic/titles/006#6-3"
    defined_for = StateCode.IN

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states["in"].tax.income
        filing_status = tax_unit("filing_status", period)
        max_deduction = p.deductions.homeowners_property_tax.max[filing_status]
        property_tax = add(tax_unit, period, ["real_estate_taxes"])
        return min_(property_tax, max_deduction)



from policyengine_us.model_api import *


class co_able_contribution_subtraction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Colorado able contribution subtraction"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://tax.colorado.gov/sites/tax/files/documents/Book0104_2023.pdf#page=14",
        "https://advance.lexis.com/documentpage/?pdmfid=1000516&crid=28da6d1b-814f-4ace-b369-4a0d0233db83&config=014FJAAyNGJkY2Y4Zi1mNjgyLTRkN2YtYmE4OS03NTYzNzYzOTg0OGEKAFBvZENhdGFsb2d592qv2Kywlf8caKqYROP5&pddocfullpath=%2Fshared%2Fdocument%2Fstatutes-legislation%2Furn%3AcontentItem%3A689J-29B3-GXF6-81G9-00008-00&pdcontentcomponentid=234176&pdteaserkey=sr2&pditab=allpods&ecomp=bs65kkk&earg=sr2&prid=a5cd01c5-3332-4547-a3ff-d2bd6bcc868f",
        # C.R.S. 39-22-104(4)(i)(II)(B)
    )
    defined_for = StateCode.CO

    def formula(tax_unit, period, parameters):
        contribution_amount = tax_unit("able_contributions", period)
        p = parameters(
            period
        ).gov.states.co.tax.income.subtractions.able_contribution
        cap_amount = p.cap[tax_unit("filing_status", period)]
        return min_(contribution_amount, cap_amount)



from policyengine_us.model_api import *


class disabled_spouse(Variable):
    value_type = bool
    entity = TaxUnit
    definition_period = YEAR
    label = "Tax unit spouse is legally disabled"

    def formula(tax_unit, period, parameters):
        person = tax_unit.members
        disabled = person("is_ssi_disabled", period)
        spouse = person("is_tax_unit_spouse", period)
        return tax_unit.any(disabled & spouse)



from policyengine_us.model_api import *


class il_tanf_demographic_eligible_person(Variable):
    value_type = bool
    entity = Person
    definition_period = MONTH
    label = "Eligible person for Illinois Temporary Assistance for Needy Families (TANF) based on demographics"
    reference = (
        "https://www.law.cornell.edu/regulations/illinois/Ill-Admin-Code-tit-89-SS-112.60",
    )
    defined_for = StateCode.IL

    def formula(person, period, parameters):
        eligible_child = person("il_tanf_eligible_child", period)
        pregnant = person("is_pregnant", period)
        return eligible_child | pregnant



from policyengine_us.model_api import *


class qualified_retirement_penalty(Variable):
    value_type = float
    entity = TaxUnit
    label = "Penalty tax on qualified retirement plans"
    unit = USD
    definition_period = YEAR



from policyengine_us.model_api import *


class me_refundable_child_care_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Maine refundable child care credit"
    unit = USD
    documentation = "Refundable portion of the Maine child care credit"
    definition_period = YEAR
    reference = [
        "https://www.mainelegislature.org/legis/statutes/36/title36sec5218.html",
        "https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_sched_a_ff.pdf#page=2",
    ]
    defined_for = StateCode.ME

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.me.tax.income.credits.child_care
        return min_(p.max_amount, tax_unit("me_child_care_credit", period))



from policyengine_us.model_api import *


class is_related_to_head_or_spouse(Variable):
    value_type = bool
    entity = Person
    label = "Person is related to head or spouse"
    definition_period = YEAR
    default_value = True



from policyengine_us.model_api import *


class ma_eaedc_if_claimed(Variable):
    value_type = float
    entity = SPMUnit
    label = "Massachusetts EAEDC benefit amount if claimed"
    unit = USD
    definition_period = MONTH
    defined_for = "ma_eaedc_eligible"
    reference = "https://www.law.cornell.edu/regulations/massachusetts/department-106-CMR/title-106-CMR-701.000"

    # Program value can not be less than 0
    # due to the eligibility requirements
    adds = ["ma_eaedc_standard_assistance"]
    subtracts = ["ma_eaedc_net_income"]



from policyengine_us.model_api import *


class de_agi_indiv(Variable):
    value_type = float
    entity = Person
    label = "Delaware adjusted gross income for each individual when married filing separately"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.DE

    def formula(person, period, parameters):
        pre_exclusions_agi = person("de_pre_exclusions_agi", period)
        indv_exclusions = person(
            "de_elderly_or_disabled_income_exclusion_indiv", period
        )
        net_income = max_(pre_exclusions_agi - indv_exclusions, 0)
        # allocate any dependent gross income to tax unit head
        is_dependent = person("is_tax_unit_dependent", period)
        sum_dep_net_income = person.tax_unit.sum(is_dependent * net_income)
        is_head = person("is_tax_unit_head", period)
        head_spouse_net_income = ~is_dependent * net_income
        head_allocated_dependent_net_income = is_head * sum_dep_net_income
        return head_spouse_net_income + head_allocated_dependent_net_income



from policyengine_us.model_api import *


class al_taxable_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "Alabama taxable income"
    defined_for = StateCode.AL
    unit = USD
    definition_period = YEAR
    reference = "https://www.revenue.alabama.gov/ultraviewer/viewer/basic_viewer/index.html?form=2023/01/22f40abk.pdf"  # 2022 Form 40A Booklet

    def formula(tax_unit, period, parameters):
        al_ded = tax_unit("al_deductions", period)
        al_agi = tax_unit("al_agi", period)
        return max_(al_agi - al_ded, 0)



from policyengine_us.model_api import *


class used_clean_vehicle_credit_potential(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Potential value of the Used clean vehicle credit"
    documentation = "Nonrefundable credit for the purchase of a previously-owned clean vehicle"
    unit = USD
    reference = "https://www.democrats.senate.gov/imo/media/doc/inflation_reduction_act_of_2022.pdf#page=370"
    defined_for = "used_clean_vehicle_credit_eligible"

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.irs.credits.clean_vehicle.used
        sale_price = tax_unit("used_clean_vehicle_sale_price", period)
        # Amount is lesser of $4,000 and 30% of sale price.
        uncapped_amount = sale_price * p.amount.percent_of_sale_price
        return min_(uncapped_amount, p.amount.max)



from policyengine_us.model_api import *


class auto_loan_interest_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Auto loan interest deduction"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.congress.gov/bill/119th-congress/house-bill/1/text"
    )

    def formula(tax_unit, period, parameters):
        auto_loan_interest = add(tax_unit, period, ["auto_loan_interest"])
        p = parameters(period).gov.irs.deductions.auto_loan_interest
        capped_interest = min_(auto_loan_interest, p.cap)
        # Get filing status.
        filing_status = tax_unit("filing_status", period)

        # Get the phaseout start amount based on filing status (line 4).
        phaseout_start = p.phase_out.start[filing_status]
        agi_pre_ald = tax_unit("adjusted_gross_income", period)
        # Get the excess amount, if any, in thousands of dollars (rounded up) [lines 5 and 6].
        excess = max_(agi_pre_ald - phaseout_start, 0)
        increments = np.ceil(excess / p.phase_out.increment)

        # Calculate the excess part phase out amount (line 7).
        phase_out_amount = increments * p.phase_out.step
        return max_(capped_interest - phase_out_amount, 0)



from policyengine_us.model_api import *


class ma_gross_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "MA gross income"
    unit = USD
    definition_period = YEAR
    reference = "https://www.mass.gov/info-details/mass-general-laws-c62-ss-2"
    defined_for = StateCode.MA

    def formula(tax_unit, period, parameters):
        # Mass. General Laws c.62 § 2(a)
        federal_gross_income = add(tax_unit, period, ["irs_gross_income"])
        foreign_earned_income = tax_unit(
            "foreign_earned_income_exclusion", period
        )
        social_security_in_agi = add(
            tax_unit, period, ["taxable_social_security"]
        )
        deductions = foreign_earned_income + social_security_in_agi
        return max_(0, federal_gross_income - deductions)



from policyengine_us.model_api import *


class ca_tanf_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Eligible for the California CalWORKs"
    definition_period = YEAR
    defined_for = StateCode.CA

    def formula(spm_unit, period, parameters):
        demographic_eligible = spm_unit("is_demographic_tanf_eligible", period)
        financial_eligibile = spm_unit("ca_tanf_financial_eligible", period)
        resources_eligible = spm_unit("ca_tanf_resources_eligible", period)
        vehicle_value_eligible = spm_unit(
            "ca_tanf_vehicle_value_eligible", period
        )

        return (
            demographic_eligible
            & financial_eligibile
            & resources_eligible
            & vehicle_value_eligible
        )



from policyengine_us.model_api import *


class nj_additions(Variable):
    value_type = float
    entity = Person
    label = "New Jersey additions to federal AGI by person"
    unit = USD
    documentation = "Additions to federal AGI to get NJ total income."
    definition_period = YEAR
    reference = "https://law.justia.com/codes/new-jersey/2022/title-54/section-54-8a-36/"
    defined_for = StateCode.NJ



from policyengine_us.model_api import *


class ny_tanf_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "New York TANF eligible"
    definition_period = YEAR
    defined_for = StateCode.NY

    def formula(spm_unit, period, parameters):
        # Assume federal demographic eligibility given consistency.
        demographic_eligible = spm_unit("is_demographic_tanf_eligible", period)
        income_eligible = spm_unit("ny_tanf_income_eligible", period)
        asset_eligible = spm_unit("ny_tanf_resources_eligible", period)
        return demographic_eligible & income_eligible & asset_eligible



from policyengine_us.model_api import *


class in_poverty(Variable):
    value_type = bool
    entity = SPMUnit
    label = "in poverty"
    documentation = "Whether a household is in poverty."
    definition_period = YEAR

    def formula(spm_unit, period, parameters):
        return spm_unit("poverty_gap", period) > 0



from policyengine_us.model_api import *


class md_total_additions(Variable):
    value_type = float
    entity = TaxUnit
    label = "MD total additions to AGI"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MD



from policyengine_us.model_api import *


class ms_agi_adjustments(Variable):
    value_type = float
    entity = Person
    label = "Mississippi adjustments to federal adjusted gross income"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.dor.ms.gov/sites/default/files/Forms/Individual/80100221.pdf#page=13",
        "https://www.dor.ms.gov/sites/default/files/Forms/Individual/80105228.pdf",  # Line 50 - 66
    )
    defined_for = StateCode.MS

    adds = "gov.states.ms.tax.income.adjustments.adjustments"



from policyengine_us.model_api import *


class nc_itemized_deductions(Variable):
    value_type = float
    entity = TaxUnit
    label = "North Carolina itemized deductions"
    unit = USD
    definition_period = YEAR
    reference = "https://www.ncdor.gov/taxes-forms/individual-income-tax/north-carolina-standard-deduction-or-north-carolina-itemized-deductions "
    defined_for = StateCode.NC

    def formula(tax_unit, period, parameters):
        # Qualified Mortgage Interest and Real Estate Property Taxes.
        filing_status = tax_unit("filing_status", period)

        mortgage_interest = add(tax_unit, period, ["mortgage_interest"])
        pirs = parameters(
            period
        ).gov.irs.deductions.itemized.salt_and_real_estate
        property_taxes = min_(
            add(tax_unit, period, ["real_estate_taxes"]),
            pirs.cap[filing_status],
        )
        pco = parameters(
            period
        ).gov.states.nc.tax.income.deductions.itemized.cap
        capped_mortage_and_property_taxes = min_(
            mortgage_interest + property_taxes, pco.mortgage_and_property_tax
        )

        # North Carolina specifies a state and local tax deduction cap which is currently not modeled in PolicyEngine

        other_deductions = add(
            tax_unit,
            period,
            ["charitable_deduction", "medical_expense_deduction"],
        )

        return capped_mortage_and_property_taxes + other_deductions



from policyengine_us.model_api import *


class ma_tcap_gross_unearned_income(Variable):
    value_type = float
    unit = USD
    entity = Person
    label = "Massachusetts Transitional Cash Assistance Program (TCAP) gross unearned income"
    definition_period = MONTH
    reference = (
        "https://www.law.cornell.edu/regulations/massachusetts/106-CMR-704-210"
    )
    defined_for = StateCode.MA

    adds = "gov.states.ma.dta.tcap.gross_income.unearned"



"""
Test script demonstrating UCGID hierarchical functionality.
"""

import numpy as np
from policyengine_us import Microsimulation
from policyengine_us.variables.household.demographic.geographic.ucgid.ucgid_enum import (
    UCGID,
)


class TestUCGIDHierarchical:
    """Test class for UCGID hierarchical functionality."""

    def test_ucgid_enum_hierarchical_methods(self):
        """Test the hierarchical methods on UCGID enum values directly."""
        print("\n=== Testing UCGID Enum Hierarchical Methods ===")

        # Test Congressional District (CA_23)
        ca_23 = UCGID.CA_23
        ca_23_codes = ca_23.get_hierarchical_codes()
        assert ca_23_codes == ["5001800US0623", "0400000US06", "0100000US"]

        # Test State (CA)
        ca = UCGID.CA
        ca_codes = ca.get_hierarchical_codes()
        assert ca_codes == ["0400000US06", "0100000US"]

        # Test US
        us = UCGID.US
        us_codes = us.get_hierarchical_codes()
        assert us_codes == ["0100000US"]

        # Test hierarchy matching
        assert ca_23.matches_hierarchy("0400000US06") == True
        assert ca_23.matches_hierarchy("0100000US") == True
        assert ca.matches_hierarchy("5001800US0623") == False

    def test_ucgid_simulation_usage(self):
        """Test how UCGID is used in a PolicyEngine simulation."""
        simulation = Microsimulation()
        ucgid_values = simulation.calculate("ucgid", "2024")

        # The UCGID value should be a string (enum name)
        ucgid_string = ucgid_values.iloc[0]  # First household
        assert type(ucgid_string) == str

        # Convert the string back to enum to access hierarchical methods
        ucgid_enum = UCGID[ucgid_string]
        hierarchical_codes = ucgid_enum.get_hierarchical_codes()

        # Should have 2 hierarchical codes for state-level UCGIDs: [state, US]
        assert len(hierarchical_codes) == 2
        assert (
            hierarchical_codes[1] == "0100000US"
        )  # Second code should always be US
        assert hierarchical_codes[0].startswith(
            "0400000US"
        )  # First should be state-level

        # Test hierarchy matching in simulation context
        is_in_us = ucgid_enum.matches_hierarchy("0100000US")
        assert is_in_us == True

    def test_ucgid_str_variable(self):
        """Test the UCGID string variable functionality."""
        simulation = Microsimulation()
        ucgid_str_values = simulation.calculate("ucgid_str", "2024")

        # The UCGID string variable should return a string representation
        ucgid_str = ucgid_str_values.iloc[0]
        assert type(ucgid_str) == str
        assert (
            ucgid_str.startswith("0100000US")
            | ucgid_str.startswith("0400000US")
            | ucgid_str.startswith("5001800US")
        )

        # Create a basic simulation to test with specific input values
        from policyengine_us import Simulation

        simulation = Simulation(
            situation={
                "people": {"person": {}},
                "households": {"household": {"members": ["person"]}},
            }
        )

        # Set a specific UCGID value for testing (CA_23)
        simulation.set_input("ucgid", 2024, UCGID.CA_23)

        # Calculate the ucgid_str value
        ucgid_str_values = simulation.calculate("ucgid_str", 2024)
        ucgid_str = ucgid_str_values[0]

        # Verify it contains all three hierarchical codes
        assert ucgid_str == "5001800US0623,0400000US06,0100000US"

    def test_ucgid_microsimulation_input_override(self):
        """Test setting UCGID input for all households in a microsimulation."""
        microsim = Microsimulation()

        # Get initial values
        ucgid_before = microsim.calculate("ucgid", 2024)
        num_households = len(ucgid_before)

        # Set all households to CA_23
        input_array = np.array([UCGID.CA_23] * num_households)
        microsim.set_input("ucgid", 2024, input_array)

        # Verify the input was set
        ucgid_after = microsim.calculate("ucgid", 2024)
        after_values = ucgid_after.values
        assert all(val == "CA_23" for val in after_values)
        assert len(set(after_values)) == 1

        # Test that ucgid_str now returns hierarchical codes for all households
        ucgid_str_values = microsim.calculate("ucgid_str", 2024)
        expected_str = "5001800US0623,0400000US06,0100000US"
        str_values = ucgid_str_values.values
        assert all(str_val == expected_str for str_val in str_values)



from policyengine_us.model_api import *


class la_general_relief_cash_asset_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    definition_period = YEAR
    label = "Eligible for the Los Angeles County General Relief based on the cash asset requirements"
    # Person has to be a resident of LA County
    defined_for = "in_la"
    reference = "https://drive.google.com/file/d/1Oc7UuRFxJj-eDwTeox92PtmRVGnG9RjW/view?usp=sharing"

    def formula(spm_unit, period, parameters):
        cash = spm_unit("spm_unit_cash_assets", period)
        limit = spm_unit("la_general_relief_cash_asset_limit", period)
        return cash <= limit



from policyengine_us.model_api import *


class hi_eitc(Variable):
    value_type = float
    entity = TaxUnit
    label = "Hawaii earned income tax credit"
    defined_for = StateCode.HI
    unit = USD
    definition_period = YEAR
    reference = "https://www.capitol.hawaii.gov/hrscurrent/Vol04_Ch0201-0257/HRS0235/HRS_0235-0055_0075.htm"

    def formula(tax_unit, period, parameters):
        federal_eitc = tax_unit("eitc", period)
        rate = parameters(period).gov.states.hi.tax.income.credits.eitc.match
        return rate * federal_eitc



from policyengine_us.model_api import *


class tax_unit_combined_income_for_social_security_taxability(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Taxable Social Security combined income"
    unit = USD
    reference = (
        "https://www.law.cornell.edu/uscode/text/26/86"
        "https://www.ssa.gov/benefits/retirement/planner/taxes.html"
    )

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.irs.social_security.taxability
        gross_ss = tax_unit("tax_unit_social_security", period)

        # The legislation directs the usage an income definition that is
        # a particularly modified AGI, plus half of gross Social Security
        # payments. We assume that the 'half' here is the same underlying
        # parameter as the lower taxability marginal rate (also 50% in the
        # baseline), and that they would be mechanically the same parameter.

        ss_fraction = p.rate.base * gross_ss
        return tax_unit("taxable_ss_magi", period) + ss_fraction



from policyengine_us.model_api import *


class traditional_403b_contributions(Variable):
    value_type = float
    entity = Person
    label = "Traditional 403(b) contributions"
    unit = USD
    documentation = "Contributions to traditional 403(b) accounts."
    definition_period = YEAR



from policyengine_us.model_api import *


class DCCCSPChildCareProviderCategory(Enum):
    CHILD_CENTER = "Child Center"
    CHILD_HOME_AND_EXPANDED_HOME = "Child Home and Expanded Home"
    IN_HOME_CHILD_CARE = "In-Home Child Care"


class dc_ccsp_childcare_provider_category(Variable):
    value_type = Enum
    entity = Person
    possible_values = DCCCSPChildCareProviderCategory
    default_value = DCCCSPChildCareProviderCategory.CHILD_CENTER
    label = "DC Child Care Subsidy Program (CCSP) child care provider category"
    definition_period = MONTH
    defined_for = StateCode.DC
    reference = "https://osse.dc.gov/sites/default/files/dc/sites/osse/publication/attachments/FY25%20Subsidy%20Reimbursement%20Rates%20English.pdf"



from policyengine_us.model_api import *


class tax_unit_childcare_expenses(Variable):
    value_type = float
    entity = TaxUnit
    label = "Childcare expenses"
    unit = USD
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        # distribute the SPM unit's childcare expenses evenly across
        # children in SPM unit's Tax units
        spm_unit = tax_unit.spm_unit
        spm_unit_childcare = spm_unit("childcare_expenses", period)
        spm_unit_count_children = add(spm_unit, period, ["is_child"])
        tax_unit_count_children = add(tax_unit, period, ["is_child"])
        # avoid array divide-by-zero warning by not using where() function
        # see the following GitHub issue for more details:
        # https://github.com/PolicyEngine/policyengine-us/issues/2494
        child_ratio = np.zeros_like(spm_unit_count_children)
        mask = spm_unit_count_children > 0
        child_ratio[mask] = (
            tax_unit_count_children[mask] / spm_unit_count_children[mask]
        )
        return spm_unit_childcare * child_ratio



from policyengine_us.model_api import *


class capped_heat_pump_heat_pump_water_heater_biomass_stove_boiler_credit(
    Variable
):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Capped credit on heat pumps, heat pump water heaters, and biomass stoves and boilers"
    documentation = "Capped credit on heat pumps, heat pump water heaters, and biomass stoves and boilers"
    unit = USD
    reference = "https://www.democrats.senate.gov/imo/media/doc/inflation_reduction_act_of_2022.pdf#page=340"

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.irs.credits.energy_efficient_home_improvement
        pre_rebate_expenditure = add(
            tax_unit,
            period,
            [
                i + "_expenditures"
                for i in p.qualified_expenditures.heat_pump_heat_pump_water_heater_biomass_stove_boiler
            ],
        )
        # NB: We assume that the credit is based on after-rebate expenditures,
        # where rebates are per-item before the total rebate cap is applied.
        REBATE_ELEMENTS = ["heat_pump", "heat_pump_water_heater"]
        rebates = add(
            tax_unit,
            period,
            ["capped_" + i + "_rebate" for i in REBATE_ELEMENTS],
        )
        post_rebate_expenditure = pre_rebate_expenditure - rebates
        rate = p.rates.property
        uncapped = post_rebate_expenditure * rate
        # Cap at either the total property cap (pre-IRA) or heat pump etc. cap (post-IRA).
        # We represent pre-IRA as an infinite heat pump cap.
        heat_pump_etc_cap = (
            p.cap.annual.heat_pump_heat_pump_water_heater_biomass_stove_boiler
        )
        if heat_pump_etc_cap == np.inf:
            cap = p.cap.annual.energy_efficient_building_property
        else:
            cap = heat_pump_etc_cap
        return min_(uncapped, cap)



from policyengine_us.model_api import *


class hi_standard_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Hawaii standard deduction"
    unit = USD
    documentation = "https://files.hawaii.gov/tax/forms/2022/n11ins.pdf"
    definition_period = YEAR
    defined_for = StateCode.HI

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.hi.tax.income.deductions.standard
        filing_status = tax_unit("filing_status", period)
        return p.amount[filing_status]



from policyengine_us.model_api import *
from microdf import MicroSeries


class household_income_decile(Variable):
    label = "household income decile"
    documentation = "Decile of household income (person-weighted)"
    entity = Household
    definition_period = YEAR
    value_type = int

    def formula(household, period, parameters):
        income = household("household_net_income", period)
        count_people = household("household_count_people", period)
        household_weight = household("household_weight", period)
        weighted_income = MicroSeries(
            income, weights=household_weight * count_people
        )
        decile = weighted_income.decile_rank().values
        # Set negatives to -1.
        # This avoids the bottom decile summing to a negative number,
        # which would flip the % change in the interface.
        return where(income < 0, -1, decile)



from policyengine_us.model_api import *


class ms_standard_deduction_indiv(Variable):
    value_type = float
    entity = Person
    label = (
        "Mississippi standard deduction when married couples file separately"
    )
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MS

    def formula(person, period, parameters):
        filing_status = person.tax_unit(
            "state_filing_status_if_married_filing_separately_on_same_return",
            period,
        )
        is_head_or_spouse = person("is_tax_unit_head_or_spouse", period)
        p = parameters(period).gov.states.ms.tax.income.deductions.standard
        return is_head_or_spouse * p.amount[filing_status]



from policyengine_us.model_api import *


class or_retirement_credit_eligible_person(Variable):
    value_type = bool
    entity = Person
    label = "Eligible person for the Oregon Retirement Income Tax Credit"
    definition_period = YEAR
    reference = "https://secure.sos.state.or.us/oard/viewSingleRule.action?ruleVrsnRsn=238290#:~:text=Eligible%20individuals%20receiving%20retirement%20pay,by%20the%20household%20income%20limitation."
    defined_for = StateCode.OR

    def formula(person, period, parameters):
        # taxpayer or spouse must be age 62+ for their pension income to count
        age = person("age", period)
        is_head = person("is_tax_unit_head", period)
        is_spouse = person("is_tax_unit_spouse", period)
        p = (
            parameters(period)
            .gov.states["or"]
            .tax.income.credits.retirement_income
        )
        return (age >= p.age_threshold) & (is_head | is_spouse)



from policyengine_us.model_api import *


class three_digit_zip_code(Variable):
    value_type = str
    entity = Household
    label = "Three-digit zipcode"
    definition_period = YEAR

    def formula(household, period, parameters):
        zip_code = np.array(household("zip_code", period)).astype(str)
        zip_code_3 = (zip_code.astype(int) // 1e2).astype(int).astype(str)
        return np.char.zfill(zip_code_3, 3)



from policyengine_us.model_api import *


class rrc_cares(Variable):
    value_type = float
    entity = TaxUnit
    label = "Recovery Rebate Credit (CARES)"
    unit = USD
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/6428"

    def formula(tax_unit, period, parameters):
        rrc = parameters(period).gov.irs.credits.recovery_rebate_credit
        filing_status = tax_unit("filing_status", period)
        agi = tax_unit("adjusted_gross_income", period)
        count_children = tax_unit("ctc_qualifying_children", period)
        # (a)(2) specifies CTC eligibility for children
        count_adults = where(tax_unit("tax_unit_is_joint", period), 2, 1)
        max_payment = (
            rrc.cares.max.adult * count_adults
            + rrc.cares.max.child * count_children
        )
        payment_reduction = rrc.cares.phase_out.rate * max_(
            0, agi - rrc.cares.phase_out.threshold[filing_status]
        )
        return max_(0, max_payment - payment_reduction)



from policyengine_us.model_api import *


class dc_gac_earned_income_after_disregard_person(Variable):
    value_type = float
    entity = Person
    label = "DC General Assistance for Children (GAC) earned income after disregard per person"
    unit = USD
    definition_period = MONTH
    reference = (
        "https://code.dccouncil.gov/us/dc/council/code/sections/4-205.05a#(e)"
    )
    defined_for = "dc_gac_eligible_child"

    def formula(person, period, parameters):
        p = parameters(period).gov.states.dc.dhs
        gross_earnings = person("dc_tanf_gross_earned_income", period)
        adjusted_income = max_(
            gross_earnings - p.gac.income_disregard.amount, 0
        )
        full_time_student = person("is_full_time_student", period)
        weekly_hours_worked = person("monthly_hours_worked", period)
        work_full_time = (
            weekly_hours_worked
            >= p.tanf.work_requirement.required_hours.single_parent.higher.amount
        )

        return where(work_full_time & ~full_time_student, adjusted_income, 0)



from policyengine_us.model_api import *


class is_aca_ptc_eligible(Variable):
    value_type = bool
    entity = Person
    label = (
        "Person is eligible for ACA premium tax credit and pays ACA premium"
    )
    definition_period = YEAR

    def formula(person, period, parameters):
        # determine status eligibility for ACA PTC
        fstatus = person.tax_unit("filing_status", period)
        separate = fstatus == fstatus.possible_values.SEPARATE
        immigration_eligible = person(
            "is_aca_ptc_immigration_status_eligible", period
        )
        taxpayer_has_itin = person.tax_unit("taxpayer_has_itin", period)
        is_status_eligible = (
            taxpayer_has_itin & ~separate & immigration_eligible
        )

        # determine coverage eligibility for ACA plan
        INELIGIBLE_COVERAGE = [
            "is_medicaid_eligible",
            "is_chip_eligible",
            "is_aca_eshi_eligible",
            "is_medicare_eligible",
        ]
        is_coverage_eligible = add(person, period, INELIGIBLE_COVERAGE) == 0

        # determine income eligibility for ACA PTC
        p = parameters(period).gov.aca
        magi_frac = person.tax_unit("aca_magi_fraction", period)
        is_income_eligible = p.ptc_income_eligibility.calc(magi_frac)

        # determine which people pay an age-based ACA plan premium
        is_aca_adult = person("age", period) > p.slcsp.max_child_age
        child_pays = person("aca_child_index", period) <= p.max_child_count
        pays_aca_premium = is_aca_adult | child_pays

        return (
            is_status_eligible
            & is_coverage_eligible
            & is_income_eligible
            & pays_aca_premium
        )



from policyengine_us.model_api import *


class wv_homestead_excess_property_tax_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "West Virginia homestead excess property tax credit"
    defined_for = "wv_homestead_excess_property_tax_credit_eligible"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://code.wvlegislature.gov/11-21-23/"
        "https://tax.wv.gov/Documents/TaxForms/2021/it140.pdf#page=13"
        "https://tax.wv.gov/Documents/PIT/2022/PersonalIncomeTaxFormsAndInstructions.2022.pdf#page=14"
    )

    def formula(tax_unit, period, parameters):
        wv_sctc = tax_unit(
            "wv_sctc", period
        )  # West Virginia homestead excess property tax credit
        property_tax = add(tax_unit, period, ["real_estate_taxes"])
        wv_ghi = tax_unit("wv_gross_household_income", period)

        p = parameters(period).gov.states.wv.tax.income.credits.heptc
        heptc_amount = max_(
            property_tax - wv_sctc - p.rate.household_income * wv_ghi, 0
        )

        return min_(heptc_amount, p.cap)



from policyengine_us.model_api import *


class nd_nonrefundable_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "North Dakota nonrefundable income tax credits"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2021-iit/form-nd-1-2021.pdf#page=2"
        "https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2021-iit/individual-income-tax-booklet-2021.pdf#page=16"
        "https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/form-nd-1-2022.pdf#page=2"
        "https://www.tax.nd.gov/sites/www/files/documents/forms/individual/2022-iit/2022-individual-income-tax-booklet.pdf#page=16"
    )
    defined_for = StateCode.ND
    adds = "gov.states.nd.tax.income.credits.nonrefundable"



from policyengine_us.model_api import *


class de_aged_personal_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Delaware aged personal credit"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://delcode.delaware.gov/title30/c011/sc02/index.html#1110"
    )
    defined_for = StateCode.DE

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.de.tax.income.credits.personal_credits
        """
          Legal code says "An additional $110 in the case of each resident
          person age 60 or over.  Tax form limits it to heads and spouses,
          not elderly dependents.
        """
        age_head = tax_unit("age_head", period)
        head_amount = p.aged.calc(age_head)

        age_spouse = tax_unit("age_spouse", period)
        spouse_amount = p.aged.calc(age_spouse)

        return head_amount + spouse_amount



from policyengine_us.model_api import *


class hi_additions(Variable):
    value_type = float
    entity = TaxUnit
    label = "Hawaii additions to federal adjusted gross income"
    reference = "https://files.hawaii.gov/tax/forms/2022/n11ins.pdf#page=11"
    defined_for = StateCode.HI
    unit = USD
    definition_period = YEAR



from policyengine_us.model_api import *


class nj_medical_expense_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "New Jersey medical expense deduction"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.NJ

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.nj.tax.income.deductions.medical_expenses
        self_employed_medical_expense_deduction = tax_unit(
            "self_employed_health_insurance_ald", period
        )
        medical_expenses = add(
            tax_unit, period, ["medical_out_of_pocket_expenses"]
        )
        agi = tax_unit("nj_agi", period)
        floor = p.rate * agi
        applicable_medical_expenses = max_(0, medical_expenses - floor)
        return (
            self_employed_medical_expense_deduction
            + applicable_medical_expenses
        )



from policyengine_us.model_api import *


class tanf_reported(Variable):
    value_type = float
    entity = Person
    definition_period = YEAR
    label = "Reported TANF"
    documentation = (
        "Amount of Temporary Assistance for Needy Families benefit reported."
    )
    unit = USD
    uprating = "gov.bls.cpi.cpi_u"



from policyengine_us.model_api import *
from policyengine_core.periods import instant


class ny_ctc_pre_2024(Variable):
    value_type = float
    entity = TaxUnit
    label = "NY CTC pre-2024 rules"
    documentation = "New York's Empire State Child Credit under pre-2024 rules (original system)"
    unit = USD
    definition_period = YEAR
    reference = "https://www.nysenate.gov/legislation/laws/TAX/606"  # (c-1)
    defined_for = "ny_ctc_pre_2024_eligible"

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.ny.tax.income.credits.ctc
        person = tax_unit.members
        age = person("age", period)
        agi = tax_unit("adjusted_gross_income", period)

        # Original CTC rules (pre-2025 and post-2027)
        if not p.pre_tcja:
            federal_ctc = tax_unit("ctc", period)
            gov = parameters(period).gov
            qualifies_for_federal_ctc = person("ctc_qualifying_child", period)
        else:
            # Initialize pre-TCJA CTC branch and parameters.
            simulation = tax_unit.simulation
            pre_tcja_ctc = simulation.get_branch("pre_tcja_ctc")
            pre_tcja_ctc.tax_benefit_system = (
                simulation.tax_benefit_system.clone()
            )
            branch_parameters = pre_tcja_ctc.tax_benefit_system.parameters
            # Update parameters to pre-TCJA values.
            for (
                ctc_parameter
            ) in branch_parameters.gov.irs.credits.ctc.get_descendants():
                if isinstance(ctc_parameter, Parameter):
                    ctc_parameter.update(
                        start=instant("2017-01-01"),
                        stop=instant("2026-01-01"),
                        value=ctc_parameter("2017-01-01"),
                    )
            # Delete all arrays from pre-TCJA CTC branch.
            for variable in pre_tcja_ctc.tax_benefit_system.variables:
                if "ctc" in variable:
                    pre_tcja_ctc.delete_arrays(variable)
            # Calculate pre-TCJA CTC.
            maximum_ctc = pre_tcja_ctc.calculate(
                "ctc_child_individual_maximum", period
            )
            meets_ny_minimum_age = age >= p.minimum_age
            pre_tcja_ctc.set_input(
                "ctc_individual_maximum",
                period,
                maximum_ctc * meets_ny_minimum_age,
            )
            max_federal_ctc = pre_tcja_ctc.tax_unit("ctc", period)
            ctc_phase_in = pre_tcja_ctc.tax_unit("ctc_phase_in", period)
            federal_ctc = min_(max_federal_ctc, ctc_phase_in)
            qualifies_for_federal_ctc = pre_tcja_ctc.person(
                "ctc_qualifying_child", period
            )
            gov = branch_parameters(period).gov

        # Remaining logic is based on NY parameters.
        qualifies = qualifies_for_federal_ctc & (age >= p.minimum_age)
        qualifying_children = tax_unit.sum(qualifies)
        federal_match = federal_ctc * p.amount.percent
        # Filers with income below the CTC phase-out threshold receive a
        # minimum amount per child.
        minimum = p.amount.minimum * qualifying_children
        federal_threshold = gov.irs.credits.ctc.phase_out.threshold[
            tax_unit("filing_status", period)
        ]
        eligible_for_minimum = agi < federal_threshold
        applicable_minimum = eligible_for_minimum * minimum
        return max_(applicable_minimum, federal_match)



from policyengine_us.model_api import *


class md_snap_is_elderly(Variable):
    value_type = bool
    entity = Person
    definition_period = YEAR
    label = "Is an elderly person for Maryland SNAP minimum allotment"
    defined_for = StateCode.MD

    def formula(person, period, parameters):
        p = parameters(period).gov.states.md.usda.snap.min_allotment
        return person("age", period) >= p.age_threshold



from policyengine_us.model_api import *


class is_co_denver_dhs_elderly(Variable):
    value_type = bool
    entity = Person
    definition_period = YEAR
    documentation = "Is elderly per Denver DHS guidelines"
    label = "Denver DHS elderly"

    def formula(person, period, parameters):
        p = parameters(period).gov.local.co.denver.dhs
        return person("age", period) >= p.elderly_age_threshold



from policyengine_us.model_api import *


class ca_foster_youth_tax_credit_person(Variable):
    value_type = float
    entity = Person
    label = "California foster youth tax credit"
    unit = USD
    definition_period = YEAR
    reference = "https://www.ftb.ca.gov/forms/2022/2022-3514.pdf#page=4"
    defined_for = "ca_foster_youth_tax_credit_eligible_person"

    def formula(person, period, parameters):
        p = parameters(period).gov.states.ca.tax.income.credits.foster_youth
        age = person("age", period)
        base_credit = p.amount.calc(age)
        earned_income = person.tax_unit("tax_unit_earned_income", period)
        excess_earned_income = max_(earned_income - p.phase_out.start, 0)
        reduction_increments = excess_earned_income / p.phase_out.increment
        reduction_amount = max_(0, reduction_increments * p.phase_out.amount)
        return max_(0, base_credit - reduction_amount)



from policyengine_us.model_api import *


class ctc_adult_individual_maximum(Variable):
    value_type = float
    entity = Person
    label = "CTC maximum amount (adult dependent)"
    unit = USD
    documentation = (
        "The CTC entitlement in respect of this person as an adult dependent."
    )
    definition_period = YEAR
    defined_for = "is_tax_unit_dependent"
    reference = (
        "https://www.law.cornell.edu/uscode/text/26/24#a",
        "https://www.law.cornell.edu/uscode/text/26/24#h",
        "https://www.law.cornell.edu/uscode/text/26/24#i",
    )

    def formula(person, period, parameters):
        p = parameters(period).gov.irs.credits.ctc
        is_adult = person("ctc_child_individual_maximum", period) == 0
        return is_adult * p.amount.adult_dependent
